from pathlib import Path

import pytest
import torch
from torch.testing import assert_allclose

from biome.text import Dataset
from biome.text import Pipeline
from biome.text import Trainer
from biome.text import TrainerConfiguration


@pytest.fixture
def pretrained_word_vectors(tmp_path) -> Path:
    file_path = tmp_path / "pretrained_word_vectors.txt"
    file_path.write_text("2 2\ntest 0.66 0.33\nthis 0.25 0.75")

    return file_path


@pytest.fixture
def dataset() -> Dataset:
    data = {"text": ["test"], "label": ["good"]}
    return Dataset.from_dict(data)


@pytest.fixture
def dataset2() -> Dataset:
    data = {"text": ["this"], "label": ["good"]}
    return Dataset.from_dict(data)


@pytest.fixture
def pipeline_config(pretrained_word_vectors) -> dict:
    config = {
        "name": "pretrained_word_vectors_test",
        "features": {
            "word": {
                "embedding_dim": 2,
                "weights_file": str(pretrained_word_vectors.absolute()),
            }
        },
        "head": {"type": "TextClassification", "labels": ["good"]},
    }
    return config


def test_create_pipeline_with_weights_file(pipeline_config, dataset, tmp_path):
    pipeline = Pipeline.from_config(pipeline_config)

    output = tmp_path / "pretrained_word_vector_output"
    trainer = Trainer(
        pipeline=pipeline,
        train_dataset=dataset,
        trainer_config=TrainerConfiguration(max_epochs=1, gpus=0),
    )
    trainer.fit(output)

    instance = pipeline.head.featurize("test")
    instance.index_fields(pipeline.vocab)

    assert_allclose(
        pipeline.backbone.embedder(instance.as_tensor_dict()["text"], 0),
        torch.tensor([[0.66, 0.33]]),
    )

    # Loading a pretrained model without the weights file should work
    Path(pipeline_config["features"]["word"]["weights_file"]).unlink()
    assert isinstance(Pipeline.from_pretrained(str(output / "model.tar.gz")), Pipeline)


def test_extending_vocab_with_weights_file(
    pipeline_config, dataset, dataset2, capsys, caplog
):
    pipeline = Pipeline.from_config(pipeline_config)
    # create vocab
    pipeline.create_vocab([dataset.to_instances(pipeline)])

    # extending the vocab with the weights file available should apply the pretrained weights
    pipeline.create_vocab([dataset2.to_instances(pipeline)])

    instance = pipeline.head.featurize("this")
    instance.index_fields(pipeline.vocab)

    assert_allclose(
        pipeline.backbone.embedder(instance.as_tensor_dict()["text"]),
        torch.tensor([[0.25, 0.75]]),
    )

    # extending the vocab with the weights file deleted should trigger a warning
    Path(pipeline_config["features"]["word"]["weights_file"]).unlink()
    ds = Dataset.from_dict({"text": ["that"], "label": ["good"]})
    pipeline.create_vocab([ds.to_instances(pipeline)])

    assert caplog.record_tuples[-1][0] == "allennlp.modules.token_embedders.embedding"
    assert caplog.record_tuples[-1][1] == 30
    assert (
        "Embedding at model_path, "
        "_head.backbone.embedder.token_embedder_word cannot locate the pretrained_file."
        in caplog.record_tuples[-1][2]
    )


def test_raise_filenotfound_error(pipeline_config, dataset):
    Path(pipeline_config["features"]["word"]["weights_file"]).unlink()
    pipeline = Pipeline.from_config(pipeline_config)

    with pytest.raises(FileNotFoundError):
        pipeline.create_vocab([dataset.to_instances(pipeline)])



import copy
import functools
import inspect
import os
import os.path
import re
from inspect import Parameter
from typing import Any
from typing import Callable
from typing import Dict
from typing import Iterable
from typing import List
from typing import Optional
from typing import Tuple
from typing import Type

import yaml
from allennlp.common import util
from allennlp.data import Token as AllenNLPToken
from allennlp.data.dataset_readers.dataset_utils import to_bioul
from spacy.tokens import Token as SpacyToken
from spacy.tokens.doc import Doc
from spacy.training import biluo_tags_to_offsets
from spacy.training import offsets_to_biluo_tags

_INVALID_TAG_CHARACTERS = re.compile(r"[^-/\w\.]")


def yaml_to_dict(filepath: str) -> Dict[str, Any]:
    """Loads a yaml file into a data dictionary

    Parameters
    ----------
    filepath
        Path to the yaml file

    Returns
    -------
    dict
    """
    with open(filepath) as yaml_content:
        config = yaml.safe_load(yaml_content)
    return config


def update_method_signature(
    signature: inspect.Signature, to_method: Callable
) -> Callable:
    """Updates the signature of a method

    Parameters
    ----------
    signature
        The signature with which to update the method
    to_method
        The method whose signature will be updated

    Returns
    -------
    updated_method
    """

    def wrapper(*args, **kwargs):
        return to_method(*args, **kwargs)

    wrapper.__signature__ = signature
    wrapper.__doc__ = to_method.__doc__

    return wrapper


def isgeneric(class_type: Type) -> bool:
    """Checks if a class type is a generic type (List[str] or Union[str, int]"""
    return hasattr(class_type, "__origin__")


def is_running_on_notebook() -> bool:
    """Checks if code is running inside a jupyter notebook"""
    try:
        import IPython

        return IPython.get_ipython().has_trait("kernel")
    except (AttributeError, NameError, ModuleNotFoundError):
        return False


def split_signature_params_by_predicate(
    signature_function: Callable, predicate: Callable
) -> Tuple[List[Parameter], List[Parameter]]:
    """Splits parameters signature by defined boolean predicate function"""
    signature = inspect.signature(signature_function)
    parameters = list(
        filter(
            lambda p: p.name != "self"
            and p.kind not in [Parameter.VAR_KEYWORD, Parameter.VAR_POSITIONAL],
            signature.parameters.values(),
        )
    )
    matches_group = list(filter(lambda p: predicate(p), parameters))
    non_matches_group = list(filter(lambda p: not predicate(p), parameters))

    return matches_group, non_matches_group


def sanitize_metric_name(name: str) -> str:
    """Sanitizes the name to comply with tensorboardX conventions when logging.

    Parameter
    ---------
    name
        Name of the metric

    Returns
    -------
    sanitized_name
    """
    if not name:
        return name
    new_name = _INVALID_TAG_CHARACTERS.sub("_", name)
    new_name = new_name.lstrip("/")
    return new_name


def save_dict_as_yaml(dictionary: dict, path: str) -> str:
    """Save a cfg dict to path as yaml

    Parameters
    ----------
    dictionary
        Dictionary to be saved
    path
        Filesystem location where the yaml file will be saved

    Returns
    -------
    path
        Location of the yaml file
    """
    dir_name = os.path.dirname(path)
    # Prevent current workdir relative routes
    # `save_dict_as_yaml("just_here.yml")
    if dir_name:
        os.makedirs(dir_name, exist_ok=True)

    with open(path, "w") as yml_file:
        yaml.dump(dictionary, yml_file, default_flow_style=False, allow_unicode=True)

    return path


def get_full_class_name(the_class: Type) -> str:
    """Given a type class return the full qualified class name """
    # o.__module__ + "." + o.__class__.__qualname__ is an example in
    # this context of H.L. Mencken's "neat, plausible, and wrong."
    # Python makes no guarantees as to whether the __module__ special
    # attribute is defined, so we take a more circumspect approach.
    # Alas, the module name is explicitly excluded from __qualname__
    # in Python 3.

    module = the_class.__module__
    if module is None or module == str.__class__.__module__:
        return the_class.__name__  # Avoid reporting __builtin__
    else:
        return module + "." + the_class.__name__


def stringify(value: Any) -> Any:
    """Creates an equivalent data structure representing data values as string

    Parameters
    ----------
    value
        Value to be stringified

    Returns
    -------
    stringified_value
    """
    if value is None or isinstance(value, str):
        return value
    if isinstance(value, dict):
        return {key: stringify(value) for key, value in value.items()}
    if isinstance(value, Iterable):
        return [stringify(v) for v in value]
    return str(value)


def sanitize_for_params(x: Any) -> Any:
    """Sanitizes the input for a more flexible usage with AllenNLP's `.from_params()` machinery.

    For now it is mainly used to transform numpy numbers to python types

    Parameters
    ----------
    x
        The parameter passed on to `allennlp.common.FromParams.from_params()`

    Returns
    -------
    sanitized_x
    """
    # AllenNLP has a similar function (allennlp.common.util.sanitize) but it does not work for my purpose, since
    # numpy types are checked only after the float type check, and:
    # isinstance(numpy.float64(1), float) == True !!!
    if isinstance(x, util.numpy.number):
        return x.item()
    elif isinstance(x, util.numpy.bool_):
        # Numpy bool_ need to be converted to python bool.
        return bool(x)
    if isinstance(x, (str, float, int, bool)):
        return x
    elif isinstance(x, dict):
        # Dicts need their values sanitized
        return {key: sanitize_for_params(value) for key, value in x.items()}
    # Lists and Tuples need their values sanitized
    elif isinstance(x, list):
        return [sanitize_for_params(x_i) for x_i in x]
    elif isinstance(x, tuple):
        return tuple(sanitize_for_params(x_i) for x_i in x)
    # We include `to_json` function customize sanitization for user defined classes
    elif hasattr(x, "to_json"):
        return x.to_json()
    return x


def sanitize_for_yaml(value: Any):
    """Sanitizes the value for a simple yaml output, that is classes only built-in types"""
    if isinstance(value, list):
        return [sanitize_for_yaml(v) for v in value]
    if isinstance(value, tuple):
        return tuple(sanitize_for_yaml(v) for v in value)
    if isinstance(value, dict):
        return {k: sanitize_for_yaml(v) for k, v in value.items()}
    try:
        yaml.dump(value)
    # we try a string representation
    except Exception:
        return str(value)
    else:
        return value


def span_labels_to_tag_labels(
    labels: List[str], label_encoding: str = "BIO"
) -> List[str]:
    """Converts a list of span labels to tag labels following `spacy.training.offsets_to_biluo_tags`

    Parameters
    ----------
    labels
        Span labels to convert
    label_encoding
        The label format used for the tag labels

    Returns
    -------
    tag_labels
    """
    if label_encoding == "BIOUL":
        converted_labels = [
            f"{char}-{label}" for char in ["B", "I", "U", "L"] for label in labels
        ] + ["O"]
    elif label_encoding == "BIO":
        converted_labels = [
            f"{char}-{label}" for char in ["B", "I"] for label in labels
        ] + ["O"]
    else:
        raise ValueError(
            f"'{label_encoding}' is not a supported label encoding scheme."
        )

    return converted_labels


def bioul_tags_to_bio_tags(tags: List[str]) -> List[str]:
    """Converts BIOUL tags to BIO tags

    Parameters
    ----------
    tags
        BIOUL tags to convert

    Returns
    -------
    bio_tags
    """
    return [tag.replace("L-", "I-", 1).replace("U-", "B-", 1) for tag in tags]


def tags_from_offsets(
    doc: Doc,
    offsets: List[Dict],
    label_encoding: Optional[str] = "BIOUL",
) -> List[str]:
    """Converts offsets to BIOUL or BIO tags using spacy's `offsets_to_biluo_tags`.

    Parameters
    ----------
    doc
        A spaCy Doc created with `text` and the backbone tokenizer
    offsets
        A list of dicts with start and end character index with respect to the doc, and the span label:
        `{"start": int, "end": int, "label": str}`
    label_encoding
        The label encoding to be used: BIOUL or BIO

    Returns
    -------
    tags (BIOUL or BIO)
    """
    tags = offsets_to_biluo_tags(
        doc, [(offset["start"], offset["end"], offset["label"]) for offset in offsets]
    )
    if label_encoding == "BIO":
        tags = bioul_tags_to_bio_tags(tags)
    return tags


def offsets_from_tags(
    doc: Doc,
    tags: List[str],
    label_encoding: Optional[str] = "BIOUL",
    only_token_spans: bool = False,
) -> List[Dict]:
    """Converts BIOUL or BIO tags to offsets

    Parameters
    ----------
    doc
        A spaCy Doc created with `text` and the backbone tokenizer
    tags
        A list of BIOUL or BIO tags
    label_encoding
        The label encoding of the tags: BIOUL or BIO
    only_token_spans
        If True, offsets contains only token index references. Default is False

    Returns
    -------
    offsets
        A list of dicts with start and end character/token index with respect to the doc and the span label:
        `{"start": int, "end": int, "start_token": int, "end_token": int, "label": str}`
    """
    # spacy's biluo_tags_to_offsets surprisingly does not check this ...
    if len(doc) != len(tags):
        raise ValueError(
            f"Number of tokens and tags must be the same, "
            f"but 'len({list(doc)}) != len({tags})"
        )

    if label_encoding == "BIO":
        tags = to_bioul(tags, encoding="BIO")

    offsets = []
    for start, end, label in biluo_tags_to_offsets(doc, tags):
        span = doc.char_span(start, end)
        data = {
            "start_token": span.start,
            "end_token": span.end,
            "label": label,
        }
        if not only_token_spans:
            data.update({"start": start, "end": end})
        offsets.append(data)
    return offsets


def merge_dicts(source: Dict[str, Any], destination: Dict[str, Any]) -> Dict[str, Any]:
    """
    Merge two dictionaries recursivelly

    Examples
    --------
    >>> a = { 'first' : { 'all_rows' : { 'pass' : 'dog', 'number' : '1' } } }
    >>> b = { 'first' : { 'all_rows' : { 'fail' : 'cat', 'number' : '5' } } }
    >>> merge_dicts(b, a)
    {'first': {'all_rows': {'pass': 'dog', 'number': '5', 'fail': 'cat'}}}

    """
    if not isinstance(destination, dict):
        return source

    result = copy.deepcopy(destination)

    for key, value in source.items():
        if isinstance(value, dict):
            # get node or create one
            node = result.setdefault(key, {})
            value = merge_dicts(value, node)
        result[key] = value

    return result


def copy_sign_and_docs(org_func):
    """Decorator to copy the signature and the docstring from the org_func"""

    def decorator(func):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            return func(*args, **kwargs)

        wrapper.__signature__ = inspect.signature(org_func)
        wrapper.__doc__ = org_func.__doc__

        return wrapper

    return decorator


def spacy_to_allennlp_token(token: SpacyToken) -> AllenNLPToken:
    return AllenNLPToken(
        text=token.text,
        idx=token.idx,
        idx_end=token.idx + len(token),
        lemma_=token.lemma_,
        pos_=token.pos_,
        tag_=token.tag_,
        dep_=token.dep_,
        ent_type_=token.ent_type_,
    )



import copy
import inspect
import json
import logging
import os
import tempfile
from inspect import Parameter
from pathlib import Path
from typing import Any
from typing import Dict
from typing import Iterable
from typing import Iterator
from typing import List
from typing import Optional
from typing import Tuple
from typing import Type
from typing import Union
from typing import cast

import mlflow
import numpy
import torch
from allennlp.common.file_utils import is_url_or_existing_file
from allennlp.data import Vocabulary
from allennlp.models import load_archive
from allennlp.models.archival import Archive
from allennlp.models.archival import archive_model

from biome.text import __version__
from biome.text import vocabulary
from biome.text.backbone import ModelBackbone
from biome.text.configuration import PipelineConfiguration
from biome.text.configuration import PredictionConfiguration
from biome.text.configuration import VocabularyConfiguration
from biome.text.dataset import Dataset
from biome.text.dataset import InstanceDataset
from biome.text.features import TransformersFeatures
from biome.text.features import WordFeatures
from biome.text.helpers import update_method_signature
from biome.text.mlflow_model import BiomeTextModel
from biome.text.model import PipelineModel
from biome.text.modules.heads import TaskHead
from biome.text.modules.heads import TaskHeadConfiguration
from biome.text.trainer import Trainer


class Pipeline:
    """Manages NLP models configuration and actions.

    Use `Pipeline` for creating new models from a configuration or loading a pretrained model.

    Use instantiated Pipelines for training from scratch, fine-tuning, predicting, serving, or exploring predictions.
    """

    _LOGGER = logging.getLogger(__name__)

    def __init__(self, model: PipelineModel, config: PipelineConfiguration):
        self._model = model
        self._config = config

        self._update_prediction_signatures()

    def _update_prediction_signatures(self):
        """Updates the `self.predict` signature to match the model inputs for interactive work-flows"""
        updated_parameters = [
            par
            for name, par in inspect.signature(self.head.featurize).parameters.items()
            if par.default == Parameter.empty
        ] + [
            par
            for name, par in inspect.signature(self.predict).parameters.items()
            if name not in ["args", "kwargs"]
        ]
        new_signature = inspect.Signature(updated_parameters)

        self.__setattr__(
            self.predict.__name__, update_method_signature(new_signature, self.predict)
        )

    @classmethod
    def from_yaml(cls, path: str, vocab_path: Optional[str] = None) -> "Pipeline":
        """Creates a pipeline from a config yaml file

        Parameters
        ----------
        path
            The path to a YAML configuration file
        vocab_path
            If provided, the pipeline vocab will be loaded from this path

        Returns
        -------
        pipeline
            A configured pipeline
        """
        pipeline_configuration = PipelineConfiguration.from_yaml(path)

        return cls.from_config(pipeline_configuration, vocab_path=vocab_path)

    @classmethod
    def from_config(
        cls,
        config: Union[PipelineConfiguration, dict],
        vocab_path: Optional[str] = None,
    ) -> "Pipeline":
        """Creates a pipeline from a `PipelineConfiguration` object or a configuration dictionary

        Parameters
        ----------
        config
            A `PipelineConfiguration` object or a configuration dict
        vocab_path
            If provided, the pipeline vocabulary will be loaded from this path

        Returns
        -------
        pipeline
            A configured pipeline
        """
        if isinstance(config, PipelineConfiguration):
            config = config.as_dict()

        model = PipelineModel(
            config=config,
            vocab=Vocabulary.from_files(vocab_path) if vocab_path is not None else None,
        )

        if not isinstance(model, PipelineModel):
            raise TypeError(f"Cannot load model. Wrong format of {model}")

        cls._add_transformers_vocab_if_needed(model)

        return cls(model, PipelineConfiguration.from_dict(config))

    @classmethod
    def from_pretrained(cls, path: Union[str, Path]) -> "Pipeline":
        """Loads a pretrained pipeline providing a *model.tar.gz* file path

        Parameters
        ----------
        path
            The path to the *model.tar.gz* file of a pretrained `Pipeline`

        Returns
        -------
        pipeline
            A pretrained pipeline
        """
        archive = load_archive(
            path,
            # Necessary for AllenNLP>=1.2.0 that requires a dataset_reader config key
            # We choose the "interleaving" type since it is the most light weight one.
            overrides={"dataset_reader": {"type": "interleaving", "readers": {}}},
        )
        model = cls._model_from_archive(archive)
        model.file_path = str(path)
        config = cls._config_from_archive(archive)

        if not isinstance(model, PipelineModel):
            raise TypeError(f"Cannot load model. Wrong format of {model}")

        return cls(model, config)

    @property
    def name(self) -> str:
        """Gets the pipeline name"""
        return self._model.name

    @property
    def inputs(self) -> List[str]:
        """Gets the pipeline input field names"""
        return self._model.inputs

    @property
    def output(self) -> List[str]:
        """Gets the pipeline output field names"""
        return self._model.output

    @property
    def backbone(self) -> ModelBackbone:
        """Gets the model backbone of the pipeline"""
        return self.head.backbone

    @property
    def head(self) -> TaskHead:
        """Gets the pipeline task head"""
        return self._model.head

    @property
    def vocab(self) -> Vocabulary:
        """Gets the pipeline vocabulary"""
        return self._model.vocab

    @property
    def config(self) -> PipelineConfiguration:
        """Gets the pipeline configuration"""
        return self._config

    @property
    def model(self) -> PipelineModel:
        """Gets the underlying model"""
        return self._model

    @property
    def type_name(self) -> str:
        """The pipeline name. Equivalent to task head name"""
        return self.head.__class__.__name__

    @property
    def num_trainable_parameters(self) -> int:
        """Number of trainable parameters present in the model.

        At training time, this number can change when freezing/unfreezing certain parameter groups.
        """
        if vocabulary.is_empty(self.vocab, self.config.features.configured_namespaces):
            self._LOGGER.warning(
                "At least one vocabulary of your features is still empty! "
                "The number of trainable parameters usually depends on the size of your vocabulary."
            )
        return sum(p.numel() for p in self._model.parameters() if p.requires_grad)

    @property
    def num_parameters(self) -> int:
        """Number of parameters present in the model."""
        if vocabulary.is_empty(self.vocab, self.config.features.configured_namespaces):
            self._LOGGER.warning(
                "At least one vocabulary of your features is still empty! "
                "The number of trainable parameters usually depends on the size of your vocabulary."
            )
        return sum(p.numel() for p in self._model.parameters())

    @property
    def named_trainable_parameters(self) -> List[str]:
        """Returns the names of the trainable parameters in the pipeline"""
        return [name for name, p in self._model.named_parameters() if p.requires_grad]

    @property
    def model_path(self) -> str:
        """Returns the file path to the serialized version of the last trained model"""
        return self._model.file_path

    def init_prediction_logger(self, output_dir: str, max_logging_size: int = 100):
        """Initializes the prediction logging.

        If initialized, all predictions will be logged to a file called *predictions.json* in the `output_dir`.

        Parameters
        ----------
        output_dir: str
            Path to the folder in which we create the *predictions.json* file.
        max_logging_size: int
            Max disk size to use for prediction logs
        """
        max_bytes = max_logging_size * 1000000
        max_bytes_per_file = 2000000
        n_backups = int(max_bytes / max_bytes_per_file)
        self._model.init_prediction_logger(
            output_dir, max_bytes=max_bytes_per_file, backup_count=n_backups
        )

    def init_prediction_cache(self, max_size: int) -> None:
        """Initializes the cache for input predictions

        Parameters
        ----------
        max_size
            Save up to max_size most recent (inputs).
        """
        self._model.init_prediction_cache(max_size)

    def create_vocab(
        self,
        instance_datasets: Iterable[InstanceDataset],
        config: Optional[VocabularyConfiguration] = None,
    ) -> Vocabulary:
        """Creates and updates the vocab of the pipeline.

        NOTE: The trainer calls this method for you. You can use this method in case you want
        to create the vocab outside of the training process.

        Parameters
        ----------
        instance_datasets
            A list of instance datasets from which to create the vocabulary.
        config
            Configurations for the vocab creation. Default: `VocabularyConfiguration()`.

        Examples
        --------
        >>> from biome.text import Pipeline, Dataset
        >>> pl = Pipeline.from_config(
        ...     {"name": "example", "head":{"type": "TextClassification", "labels": ["pos", "neg"]}}
        ... )
        >>> dataset = Dataset.from_dict({"text": ["Just an example"], "label": ["pos"]})
        >>> instance_dataset = dataset.to_instances(pl)
        >>> vocab = pl.create_vocab([instance_dataset])
        """
        # The transformers feature comes with its own vocab, no need to create anything if it is the only feature
        if self.config.features.configured_namespaces == [
            TransformersFeatures.namespace
        ]:
            return self.vocab

        self._check_for_word_vector_weights_file()

        config = config or VocabularyConfiguration()

        vocab = Vocabulary.from_instances(
            instances=(
                instance for dataset in instance_datasets for instance in dataset
            ),
            max_vocab_size=config.max_vocab_size,
            min_count=config.min_count,
            pretrained_files=config.pretrained_files,
            only_include_pretrained_words=config.only_include_pretrained_words,
            min_pretrained_embeddings=config.min_pretrained_embeddings,
            tokens_to_add=config.tokens_to_add,
        )

        # If the vocab is the same, this is just a no-op
        self._model.extend_vocabulary(vocab)

        return vocab

    def _check_for_word_vector_weights_file(self):
        # If the vocab is empty, we assume this is an untrained pipeline
        # and we want to raise an error if the weights file is not found.
        # Extending the vocab with a non-existent weights file only throws a warning.
        try:
            assert is_url_or_existing_file(Path(self.config.features.word.weights_file))
        except AssertionError:
            if vocabulary.is_empty(self.vocab, [WordFeatures.namespace]):
                raise FileNotFoundError(
                    f"Cannot find the weights file {self.config.features.word.weights_file}"
                )
        # no word feature, or weights_file is None
        except (AttributeError, TypeError):
            pass

    def _restore_vocab(self, folder: str) -> Vocabulary:
        # The transformers feature comes with its own vocab, no need to restore anything if it is the only feature
        if self.config.features.configured_namespaces == [
            TransformersFeatures.namespace
        ]:
            return self.vocab

        self._check_for_word_vector_weights_file()

        vocab = Vocabulary.from_files(folder)
        self._model.extend_vocabulary(vocab)

        return vocab

    def predict(
        self,
        *args,
        batch: Optional[List[Dict[str, Any]]] = None,
        add_tokens: bool = False,
        add_attributions: bool = False,
        attributions_kwargs: Optional[Dict] = None,
        **kwargs,
    ) -> Union[Dict[str, numpy.ndarray], List[Optional[Dict[str, numpy.ndarray]]]]:
        """Returns a prediction given some input data based on the current state of the model

        The accepted input is dynamically calculated and can be checked via the `self.inputs` attribute
        (`print(Pipeline.inputs)`)

        Parameters
        ----------
        *args/**kwargs
            These are dynamically updated and correspond to the pipeline's `self.inputs`.
            If provided, the `batch` parameter will be ignored.
        batch
            A list of dictionaries that represents a batch of inputs. The dictionary keys must comply with the
            `self.inputs` attribute. Predicting batches should typically be faster than repeated calls with args/kwargs.
        add_tokens
            If true, adds a 'tokens' key in the prediction that contains the tokenized input.
        add_attributions
            If true, adds a 'attributions' key that contains attributions of the input to the prediction.
        attributions_kwargs
            This dict is directly passed on to the `TaskHead.compute_attributions()`.

        Returns
        -------
        predictions
            A dictionary or a list of dictionaries containing the predictions and additional information.
            If a prediction fails for a single input in the batch, its return value will be `None`.

        Raises
        ------
        PredictionError
            Failed to predict the single input or the whole batch
        """
        if args or kwargs:
            batch = [self._map_args_kwargs_to_input(*args, **kwargs)]

        prediction_config = PredictionConfiguration(
            add_tokens=add_tokens,
            add_attributions=add_attributions,
            attributions_kwargs=attributions_kwargs or {},
        )

        predictions = self._model.predict(batch, prediction_config)
        if not any(predictions):
            raise PredictionError(f"Failed to make predictions for {batch}")

        predictions_dict = [
            prediction.as_dict() if prediction is not None else None
            for prediction in predictions
        ]

        return predictions_dict[0] if (args or kwargs) else predictions_dict

    def _map_args_kwargs_to_input(self, *args, **kwargs) -> Dict[str, Any]:
        """Helper function for the `self.predict` method"""
        input_dict = {k: v for k, v in zip(self.inputs, args)}
        input_dict.update(kwargs)

        return input_dict

    def evaluate(
        self,
        test_dataset: Union[Dataset, InstanceDataset],
        batch_size: int = 16,
        lazy: bool = False,
        output_dir: Optional[Union[str, Path]] = None,
        verbose: bool = True,
    ) -> Dict[str, Any]:
        """Evaluate your model on a test dataset

        Parameters
        ----------
        test_dataset
            The test data set.
        batch_size
            The batch size. Default: 16.
        lazy
            If True, instances are lazily loaded from disk, otherwise they are loaded into memory.
            Ignored when `test_dataset` is a `InstanceDataset`. Default: False.
        output_dir
            Save a `metrics.json` to this output directory. Default: None.
        verbose
            If True, prints the test results. Default: True.

        Returns
        -------
        Dict[str, Any]
            A dictionary with the metrics
        """
        trainer = Trainer(self, lazy=lazy)

        return trainer.test(
            test_dataset, batch_size=batch_size, output_dir=output_dir, verbose=verbose
        )

    def set_head(self, type: Type[TaskHead], **kwargs):
        """Sets a new task head for the pipeline

        Call this to reuse the weights and config of a pre-trained model (e.g., language model) for a new task.

        Parameters
        ----------
        type: `Type[TaskHead]`
            The `TaskHead` class to be set for the pipeline (e.g., `TextClassification`
        **kwargs:
            The `TaskHead` specific arguments (e.g., the classification head needs a `pooler` layer)
        """

        self._config.head = TaskHeadConfiguration(type=type, **kwargs)
        self._model.set_head(self._config.head.compile(backbone=self.backbone))

    def model_parameters(self) -> Iterator[Tuple[str, torch.Tensor]]:
        """Returns an iterator over all model parameters, yielding the name and the parameter itself.

        Examples
        --------
        You can use this to freeze certain parameters in the training:

        >>> pipeline = Pipeline.from_config({
        ...     "name": "model_parameters_example",
        ...     "head": {"type": "TextClassification", "labels": ["a", "b"]},
        ... })
        >>> for name, parameter in pipeline.model_parameters():
        ...     if not name.endswith("bias"):
        ...         parameter.requires_grad = False

        """
        return self._model.named_parameters()

    def copy(self) -> "Pipeline":
        """Returns a copy of the pipeline"""
        model = PipelineModel(self._config.as_dict(), vocab=copy.deepcopy(self.vocab))
        config = copy.deepcopy(self._config)

        pipeline_copy = Pipeline(model, config)
        pipeline_copy._model.load_state_dict(self._model.state_dict())

        return pipeline_copy

    def save(self, directory: Union[str, Path]) -> str:
        """Saves the pipeline in the given directory as `model.tar.gz` file.

        Parameters
        ----------
        directory
            Save the 'model.tar.gz' file to this directory.

        Returns
        -------
        file_path
            Path to the 'model.tar.gz' file.
        """
        if isinstance(directory, str):
            directory = Path(directory)

        directory.mkdir(exist_ok=True)
        with tempfile.TemporaryDirectory() as temp_dir:
            temp_path = Path(temp_dir)
            self.vocab.save_to_files(str(temp_path / "vocabulary"))
            torch.save(self._model.state_dict(), temp_path / "best.th")
            with (temp_path / "config.json").open("w") as file:
                json.dump(
                    {
                        "model": {
                            "config": self.config.as_dict(),
                            "type": "PipelineModel",
                        }
                    },
                    file,
                    indent=4,
                )
            archive_model(temp_path, archive_path=directory)

        return str(directory / "model.tar.gz")

    def to_mlflow(
        self,
        tracking_uri: Optional[str] = None,
        experiment_id: Optional[int] = None,
        run_name: str = "log_biometext_model",
        input_example: Optional[Dict] = None,
        conda_env: Optional[Dict] = None,
    ) -> str:
        """Logs the pipeline as MLFlow Model to a MLFlow Tracking server

        Parameters
        ----------
        tracking_uri
            The URI of the MLFlow tracking server, MLFlow defaults to './mlruns'. Default: None
        experiment_id
            ID of the experiment under which to create the logging run. If this argument is unspecified,
            will look for valid experiment in the following order: activated using `mlflow.set_experiment`,
            `MLFLOW_EXPERIMENT_NAME` environment variable, `MLFLOW_EXPERIMENT_ID` environment variable,
            or the default experiment as defined by the tracking server.
        run_name
            The name of the MLFlow run logging the model. Default: 'log_biometext_model'.
        input_example
            You can provide an input example in the form of a dictionary. For example, for a TextClassification head
            this would be `{"text": "This is an input example"}`.
        conda_env
            This conda environment is used when serving the model via `mlflow models serve`. Default:
            conda_env = {
                "name": "mlflow-dev",
                "channels": ["defaults", "conda-forge"],
                "dependencies": ["python=3.7.9", "pip>=20.3.0", {"pip": ["biome-text=={__version__}"]}],
            }

        Returns
        -------
        model_uri
            The URI of the logged MLFlow model. The model gets logged as an artifact to the corresponding run.

        Examples
        --------
        After logging the pipeline to MLFlow you can use the MLFlow model for inference:
        >>> import mlflow, pandas, biome.text
        >>> pipeline = biome.text.Pipeline.from_config({
        ...     "name": "to_mlflow_example",
        ...     "head": {"type": "TextClassification", "labels": ["a", "b"]},
        ... })
        >>> model_uri = pipeline.to_mlflow()
        >>> model = mlflow.pyfunc.load_model(model_uri)
        >>> prediction: pandas.DataFrame = model.predict(pandas.DataFrame([{"text": "Test this text"}]))
        """
        if tracking_uri:
            mlflow.set_tracking_uri(tracking_uri)

        # This conda environment is only needed when serving the model later on with `mlflow models serve`
        conda_env = conda_env or {
            "name": "mlflow-dev",
            "channels": ["defaults", "conda-forge"],
            "dependencies": [
                "python=3.7.9",
                "pip>=20.3.0",
                {"pip": [f"biome-text=={__version__}"]},
            ],
        }

        with tempfile.TemporaryDirectory() as tmpdir_name:
            file_path = Path(self.save(directory=tmpdir_name))

            with mlflow.start_run(
                experiment_id=experiment_id, run_name=run_name
            ) as run:
                mlflow.log_artifact(str(file_path), "biometext_pipeline")
                mlflow.pyfunc.log_model(
                    artifact_path="mlflow_model",
                    python_model=BiomeTextModel(),
                    artifacts={
                        BiomeTextModel.ARTIFACT_CONTEXT: mlflow.get_artifact_uri(
                            f"biometext_pipeline/{file_path.name}"
                        )
                    },
                    input_example=input_example,
                    conda_env=conda_env,
                )
                model_uri = os.path.join(run.info.artifact_uri, "mlflow_model")

        return model_uri

    @staticmethod
    def _add_transformers_vocab_if_needed(model: PipelineModel):
        """Adds the transformers vocabulary to the `vocab`

        Parameters
        ----------
        vocab
            The transformers vocabulary will be added to this vocab
        """
        # The AllenNLP`s PretrainedTransformerIndexer adds its specific vocabulary to the Model's vocab
        # when the first `tokens_to_index()` is called via the private _add_encoding_to_vocabulary_if_needed method.
        # We trigger this here manually in a super ugly way ...
        # Actually i am not sure why they add it to their vocab in the first place ...
        transformers_indexer = model.head.backbone.featurizer.indexer.get(
            TransformersFeatures.namespace
        )
        if transformers_indexer is not None:
            try:
                transformers_indexer._add_encoding_to_vocabulary_if_needed(model.vocab)
            except AttributeError:
                transformers_indexer._matched_indexer._add_encoding_to_vocabulary_if_needed(
                    model.vocab
                )

    @staticmethod
    def _model_from_archive(archive: Archive) -> PipelineModel:
        if not isinstance(archive.model, PipelineModel):
            raise ValueError(f"Wrong pipeline model: {archive.model}")
        return cast(PipelineModel, archive.model)

    @staticmethod
    def _config_from_archive(archive: Archive) -> PipelineConfiguration:
        config = archive.config["model"]["config"]
        return PipelineConfiguration.from_params(config)


class PredictionError(Exception):
    """Exception for a failed prediction of a single input or a whole batch"""

    # For now this is the only Error in this module. If you want to add further Errors
    # define a base class as recommended in https://docs.python.org/3/tutorial/errors.html#user-defined-exceptions
    pass



import pytest
from allennlp.data import Batch
from numpy.testing import assert_allclose

from biome.text import Pipeline
from biome.text.modules.heads.task_prediction import Attribution
from biome.text.modules.heads.task_prediction import TextClassificationPrediction


@pytest.fixture
def pipeline() -> Pipeline:
    labels = ["a", "b", "c", "d", "e", "f"]
    return Pipeline.from_config(
        {
            "name": "test_text_classification",
            "head": {"type": "TextClassification", "labels": labels},
        }
    )


def test_make_task_prediction(pipeline):
    instance = pipeline.head.featurize("test this sentence")
    forward_output = pipeline._model.forward_on_instances([instance])

    prediction = pipeline.head._make_task_prediction(forward_output[0], None)

    assert isinstance(prediction, TextClassificationPrediction)
    assert isinstance(prediction.labels, list) and isinstance(
        prediction.probabilities, list
    )
    assert len(prediction.labels) == len(prediction.probabilities) == 6
    # check descending order
    assert_allclose(
        sorted(prediction.probabilities, reverse=True), prediction.probabilities
    )
    assert all([isinstance(label, str) for label in prediction.labels])
    assert set(pipeline.head.labels) == set(prediction.labels)
    assert all([isinstance(prob, float) for prob in prediction.probabilities])


def test_compute_attributions(pipeline):
    instance = pipeline.head.featurize("test this sentence")
    pipeline.model.eval()
    forward_output = pipeline.model.forward_on_instances([instance])

    attributions = pipeline.head._compute_attributions(
        forward_output[0], instance, n_steps=1
    )

    assert all([isinstance(attribution, Attribution) for attribution in attributions])
    assert len(attributions) == 3
    assert all([attr.field == "text" for attr in attributions])
    assert all([isinstance(attr.attribution, float) for attr in attributions])
    assert attributions[1].start == 5 and attributions[1].end == 9


def test_metrics(pipeline):
    instance = pipeline.head.featurize(text="test this", label="a")
    batch = Batch([instance])
    batch.index_instances(pipeline.vocab)

    pipeline.head.forward(**batch.as_tensor_dict())
    # validation metric should have never been called
    assert pipeline.head._metrics.get_dict()["accuracy"].total_count == 1
    assert pipeline.head._metrics.get_dict(is_train=False)["accuracy"].total_count == 0

    train_metrics = pipeline.head.get_metrics(reset=True)
    expected_metric_names = (
        ["accuracy"]
        + [
            f"{label}/{metric}"
            for label in ["micro", "macro"]
            for metric in ["precision", "recall", "fscore"]
        ]
        + [
            f"_{metric}/{label}"
            for metric in ["precision", "recall", "fscore"]
            for label in ["a", "b", "c", "d", "e", "f"]
        ]
    )
    assert all(name in train_metrics for name in expected_metric_names)

    pipeline.head.training = False
    pipeline.head.forward(**batch.as_tensor_dict())
    # training metric should have never been called after its reset
    assert pipeline.head._metrics.get_dict()["accuracy"].total_count == 0
    assert pipeline.head._metrics.get_dict(is_train=False)["accuracy"].total_count == 1

    valid_metrics = pipeline.head.get_metrics()
    assert all(name in valid_metrics for name in expected_metric_names)



from typing import Any
from typing import Dict
from typing import List
from typing import Union

from allennlp.data import Instance
from allennlp.data import Token
from allennlp.data import TokenIndexer
from allennlp.data.fields import ListField
from allennlp.data.fields import TextField
from spacy.tokens import Token as SpacyToken

from biome.text.features import WordFeatures
from biome.text.tokenizer import Tokenizer


class InputFeaturizer:
    """Transforms input text (words and/or characters) into indexes and embedding vectors.

    This class defines two input features, words and chars for embeddings at word and character level respectively.

    You can provide additional features by manually specify `indexer` and `embedder` configurations within each
    input feature.

    Attributes
    ----------
    tokenizer : `Tokenizer`
        Tokenizes the input depending on its type (str, List[str], Dict[str, Any])
    indexer : `Dict[str, TokenIdexer]`
        Features dictionary for token indexing. Built from `FeaturesConfiguration`
    """

    def __init__(
        self,
        tokenizer: Tokenizer,
        indexer: Dict[str, TokenIndexer],
    ):
        self.tokenizer = tokenizer
        self.indexer = indexer

    def __call__(
        self,
        data: Union[str, List[str], List[Token], Dict[str, str]],
        to_field: str = "record",
        aggregate: bool = False,
        tokenize: bool = True,
        exclude_record_keys: bool = False,
    ) -> Instance:
        """Generates an allennlp instance from a record input.

        Parameters
        ----------
        data
            Input data
        to_field
            The field name in the returned instance
        aggregate
            If true, the returned instance will contain a single `TextField` with all record fields;
            If false, the instance will contain a `ListField` of `TextField`s.
        tokenize
            If false, skip tokenization phase and pass record data as tokenized token list.
        exclude_record_keys
            If true, excludes record keys from the output tokens in the featurization of dictionaries

        Returns
        -------
        instance

        Raises
        ------
        FeaturizeError
            We cannot featurize the input
        """
        if self._contains_empty_strings(data):
            raise FeaturizeError(
                f"The provided input data contains empty strings/tokens: {data}"
            )

        if tokenize:
            record_tokens = self._data_tokens(data, exclude_record_keys)
        else:
            record_tokens = [[t if isinstance(t, Token) else Token(t) for t in data]]

        if self._contains_empty_strings(record_tokens):
            raise FeaturizeError(
                f"Empty tokens are produced for the provided input data: {data}"
            )

        return Instance({to_field: self._tokens_to_field(record_tokens, aggregate)})

    def _contains_empty_strings(
        self,
        data: Union[
            str,
            Union[Token, SpacyToken],
            List[Union[str, Token, SpacyToken]],
            List[List[Union[str, Token, SpacyToken]]],
            Dict[str, str],
        ],
    ) -> bool:
        """Checks if the data contains empty strings/tokens

        Parameters
        ----------
        data
            The provided input data

        Returns
        -------
        bool

        Raises
        ------
        FeaturizeError
            If the input data contains a wrong data type
        """
        if isinstance(data, list):
            return (
                any([self._contains_empty_strings(item) for item in data])
                # We want to return True for empty lists
                or not data
            )
        if isinstance(data, dict):
            return (
                any([self._contains_empty_strings(item) for item in data.values()])
                # We want to return True for empty dicts
                or not data
            )
        if isinstance(data, str):
            return data == ""
        if isinstance(data, (Token, SpacyToken)):
            return data.text == ""

        raise FeaturizeError(
            f"Wrong input type for the featurizer: {(type(data), data)}"
        )

    def _data_tokens(
        self, data: Union[str, List[str], Dict[str, Any]], exclude_record_keys: bool
    ) -> List[List[Token]]:
        """Converts the input data into a list of list of tokens depending on its type.

        Parameters
        ----------
        data: Union[str, List[str], Dict[str, Any]]
            The input data
        exclude_record_keys: `bool`
            If true, excludes record keys from the output tokens in the featurization of dictionaries

        Returns
        -------
        list_of_list_of_tokens
        """
        if isinstance(data, dict):
            return self.tokenizer.tokenize_record(data, exclude_record_keys)
        if isinstance(data, str):
            return self.tokenizer.tokenize_text(data)
        return self.tokenizer.tokenize_document(data)

    def _tokens_to_field(
        self, tokens: List[List[Token]], aggregate: bool
    ) -> Union[ListField, TextField]:
        """Wraps the tokens with the indexer in one or several `TextField`s

        Parameters
        ----------
        tokens: List[List[Token]]
            The list of list of input tokens
        aggregate: `bool`
            If true, the returned instance will contain a single `TextField` with all record fields;
            If false, the instance will contain a `ListField` of `TextField`s.

        Returns
        -------
        field
            `TextField` if aggregate is true, `ListField` otherwise
        """
        if aggregate:
            return TextField(
                [token for entry_tokens in tokens for token in entry_tokens],
                self.indexer,
            )
        return ListField(
            [TextField(entry_tokens, self.indexer) for entry_tokens in tokens]
        )

    @property
    def has_word_features(self) -> bool:
        """Checks if word features are already configured as part of the featurization"""
        return WordFeatures.namespace in self.indexer


class FeaturizeError(Exception):
    """Base class for exceptions in this module"""

    pass



import os
from pathlib import Path

import pytest


def pytest_configure(config):
    # It's really hard to do testing with wandb enabled ...
    os.environ["WANDB_MODE"] = "disabled"


@pytest.fixture
def resources_path() -> Path:
    return Path(__file__).parent / "resources"


@pytest.fixture
def resources_data_path(resources_path) -> Path:
    return resources_path / "data"


@pytest.fixture
def tutorials_path() -> Path:
    repo_root = Path(__file__).parent.parent
    return repo_root / "docs" / "docs" / "documentation" / "tutorials"


@pytest.fixture
def configurations_path() -> Path:
    repo_root = Path(__file__).parent.parent
    return (
        repo_root
        / "docs"
        / "docs"
        / "documentation"
        / "user-guides"
        / "2-configuration.md"
    )


@pytest.fixture
def change_to_tmp_working_dir(tmp_path) -> Path:
    cwd = os.getcwd()
    os.chdir(tmp_path)
    yield tmp_path
    os.chdir(cwd)



from pathlib import Path

import pytest
from numpy.testing import assert_allclose

from biome.text import Dataset
from biome.text import Pipeline
from biome.text import Trainer
from biome.text import TrainerConfiguration
from biome.text.features import TransformersFeatures


@pytest.fixture
def train_dataset() -> Dataset:
    """Creates the training dataset"""
    source = (
        Path(__file__).parent.parent
        / "resources"
        / "data"
        / "emotions_with_transformers.txt"
    )

    train_dataset = Dataset.from_csv(
        paths=str(source), delimiter=";", column_names=["text", "label"]
    )
    return train_dataset


@pytest.fixture
def pipeline_dict() -> dict:
    """Creation of pipeline dictionary"""

    pipeline_dict = {
        "name": "emotions_with_transformers",
        "features": {
            "transformers": {"model_name": "sshleifer/tiny-distilbert-base-cased"}
        },
        "head": {
            "type": "TextClassification",
            "labels": [
                "anger",
                "fear",
                "joy",
                "love",
                "sadness",
                "surprise",
            ],
            "pooler": {
                "type": "bert_pooler",
                "pretrained_model": "sshleifer/tiny-distilbert-base-cased",
                "requires_grad": True,
                "dropout": 0.1,
            },
        },
    }

    return pipeline_dict


@pytest.fixture
def trainer_config(tmp_path) -> TrainerConfiguration:
    return TrainerConfiguration(
        batch_size=16,
        max_epochs=1,
        optimizer={
            "type": "adam",
            "lr": 0.0001,
        },
        gpus=0,
        default_root_dir=str(tmp_path),
    )


def test_pure_transformers(tmp_path, pipeline_dict, trainer_config, train_dataset):
    """Testing a Transformer training process and a model load"""

    pl = Pipeline.from_config(pipeline_dict)

    # Check a fixed vocabulary size for the model
    assert pl.backbone.vocab.get_vocab_size("transformers") == 28996

    pl.predict(text="test")

    output = tmp_path / "output"
    trainer = Trainer(
        pipeline=pl, train_dataset=train_dataset, trainer_config=trainer_config
    )
    trainer.fit(output_dir=output)

    # Test vocabulary from a pretrained file
    pl = Pipeline.from_pretrained(str(output / "model.tar.gz"))

    # Check a fixed vocabulary size for the model after loading
    assert pl.backbone.vocab.get_vocab_size("transformers") == 28996


def test_transformers_and_word(tmp_path, pipeline_dict, trainer_config, train_dataset):
    """Testing Transformer pipeline with an added word feature layer"""
    # Changing the pipeline to delete the BERT pooler and add a word feature
    del pipeline_dict["head"]["pooler"]
    pipeline_dict["features"].update(
        {"word": {"embedding_dim": 16, "lowercase_tokens": True}}
    )

    pl = Pipeline.from_config(pipeline_dict)
    pl.predict(text="test")

    output = tmp_path / "output"
    trainer = Trainer(
        pipeline=pl, train_dataset=train_dataset, trainer_config=trainer_config
    )
    trainer.fit(output_dir=output)

    # Check a fixed vocabulary size for the transformer and the word feature
    assert pl.backbone.vocab.get_vocab_size("transformers") == 28996
    assert pl.backbone.vocab.get_vocab_size("word") == 273

    # Test vocab from a pretrained file
    pl = Pipeline.from_pretrained(str(output / "model.tar.gz"))

    # Check a fixed vocabulary size for the transformer and the word feature after loading
    assert pl.backbone.vocab.get_vocab_size("transformers") == 28996
    assert pl.backbone.vocab.get_vocab_size("word") == 273


def test_max_length_not_affecting_shorter_sequences(pipeline_dict):
    """Max length change should not affect at all previous shorter-length models"""

    pl = Pipeline.from_config(pipeline_dict)
    state_dict = pl._model.state_dict()  # dict with the whole state of the module
    probs = pl.predict("Test this")["probabilities"]  # probabilities of the test input

    pipeline_dict["features"]["transformers"]["max_length"] = 100  # changing max length
    pl = Pipeline.from_config(pipeline_dict)
    pl._model.load_state_dict(state_dict)  # loading previous state from dict
    probs_max_length = pl.predict("Test this")["probabilities"]

    assert_allclose(probs, probs_max_length)


def test_serialization(pipeline_dict):
    """Testing object saving. Model from the pipeline must be equal to the model from .json"""

    feature = TransformersFeatures(**pipeline_dict["features"]["transformers"])
    assert feature == TransformersFeatures(**feature.to_json())



import copy
import inspect
from typing import Any
from typing import Dict
from typing import Generic
from typing import Optional
from typing import Type
from typing import TypeVar
from typing import Union

from allennlp.common import FromParams
from allennlp.common import Params
from allennlp.modules.bimpm_matching import BiMpmMatching
from allennlp.modules.seq2seq_encoders import PytorchSeq2SeqWrapper
from allennlp.modules.seq2vec_encoders import PytorchSeq2VecWrapper

from biome.text import helpers


def _find_input_attribute(component: Type[Any]) -> str:
    """Find the properly input dimension attribute name for a given component"""
    input_dim_attribute = None
    if issubclass(component, (PytorchSeq2SeqWrapper, PytorchSeq2VecWrapper)):
        input_dim_attribute = "input_size"
    elif component is BiMpmMatching:
        input_dim_attribute = "hidden_dim"
    else:
        init_method_keys = inspect.signature(component.__init__).parameters.keys()
        for param_name in ["embedding_dim", "input_dim"]:
            if param_name in init_method_keys:
                input_dim_attribute = param_name
                break
    return input_dim_attribute


T = TypeVar("T")


class ComponentConfiguration(Generic[T], FromParams):
    """
    The layer spec component allows create Pytorch modules lazily,
    and instantiate them inside a context (Model or other component) dimension layer chain.

    The layer spec wraps a component params and will generate an instance of type T once the input_dim is set.

    """

    @classmethod
    def from_params(cls: Type[T], params: Params, **extras) -> T:
        return cls(**params.as_dict())

    def __resolve_layer_class(
        self, type_name: Optional[Union[Type, str]] = None
    ) -> Type[T]:
        if isinstance(type_name, Type):
            return type_name

        layer_class = getattr(self.__class__, "__orig_bases__")[0].__args__[0]
        return layer_class.by_name(type_name) if type_name else layer_class

    def __init__(self, **config):
        self._layer_class = self.__resolve_layer_class(config.get("type"))
        config["type"] = helpers.get_full_class_name(self._layer_class)
        self._config = config or {}

    def input_dim(self, input_dim: int) -> "ComponentConfiguration":
        """Sets the input dimension attribute for this layer configuration"""
        self.__update_config_with_input_dim(input_dim)
        return self

    def __update_config_with_input_dim(self, input_dim: int):
        input_dim_attribute = _find_input_attribute(self._layer_class)

        if input_dim_attribute:
            self._config[input_dim_attribute] = input_dim

    @property
    def config(self) -> Dict[str, Any]:
        """Component read-only configuration"""
        return copy.deepcopy(self._config)

    def compile(self, **extras) -> T:
        """
        Using the wrapped configuration and the input dimension, generates a
        instance of type T representing the layer configuration
        """
        if not self.config:
            raise ValueError(f"No configuration found for {self}")

        config = self.config
        if "type" in config:
            config.pop("type")

        return self._layer_class.from_params(Params(config), **extras)



import copy
from typing import TYPE_CHECKING
from typing import Any
from typing import Dict
from typing import Iterable
from typing import List

import spacy
from allennlp.common import Params
from allennlp.common.util import get_spacy_model
from allennlp.data import Token
from allennlp.data.tokenizers import PretrainedTransformerTokenizer
from spacy.language import Language
from spacy.tokens.doc import Doc

from biome.text.helpers import spacy_to_allennlp_token
from biome.text.text_cleaning import TextCleaning

if TYPE_CHECKING:
    from biome.text.configuration import TokenizerConfiguration


class Tokenizer:
    """Pre-processes and tokenizes the input text

    Transforms inputs (e.g., a text, a list of texts, etc.) into structures containing `allennlp.data.Token` objects.

    Use its arguments to configure the first stage of the pipeline (i.e., pre-processing a given set of text inputs.)

    Use methods for tokenization depending on the shape of the inputs
    (e.g., records with multiple fields, sentences lists).

    Parameters
    ----------
    config
        A `TokenizerConfiguration` object
    """

    __SPACY_SENTENCIZER__ = "sentencizer"

    def __init__(self, config: "TokenizerConfiguration"):
        self._config = config

        self._end_tokens = config.end_tokens or []
        self._start_tokens = config.start_tokens or []
        # We reverse the tokens here because we're going to insert them with `insert(0)` later;
        # this makes sure they show up in the right order.
        self._start_tokens.reverse()

        self.__nlp__ = get_spacy_model(
            self.config.spacy_model, pos_tags=True, ner=False, parse=False
        )
        if config.segment_sentences and not self.__nlp__.has_pipe(
            self.__SPACY_SENTENCIZER__
        ):
            self.__nlp__.add_pipe(self.__SPACY_SENTENCIZER__)

        if config.text_cleaning is None:
            self.text_cleaning = TextCleaning()
        else:
            self.text_cleaning = TextCleaning.from_params(
                Params(copy.deepcopy(config.text_cleaning))
            )

    @property
    def config(self) -> "TokenizerConfiguration":
        return self._config

    @property
    def nlp(self) -> Language:
        return self.__nlp__

    def tokenize_text(self, text: str) -> List[List[Token]]:
        """
        Tokenizes a text string applying sentence segmentation, if enabled

        Parameters
        ----------
        text: `str`
            The input text

        Returns
        -------
        A list of list of `Token`.

        If no sentence segmentation is enabled, or just one sentence is found in text
        the first level list will contain just one element: the tokenized text.

        """
        return self.tokenize_document([text])

    def _tokenize(self, text: str) -> List[Token]:
        """Tokenizes an input text string

        The simplest case where your input is just a `str`. For a text tokenization with
        sentence segmentation, see `tokenize_text`


        Parameters
        ----------
        text: `str`
            The input text
        Returns
        -------
        tokens: `List[Token]`

        """
        tokens = self.nlp(text[: self.config.max_sequence_length])
        if self.config.remove_space_tokens:
            tokens = [token for token in tokens if not token.is_space]

        return self._sanitize(tokens)

    def tokenize_document(self, document: List[str]) -> List[List[Token]]:
        """Tokenizes a document-like structure containing lists of text inputs

        Use this to account for hierarchical text structures (e.g., a paragraph)

        Parameters
        ----------
        document: `List[str]`
            A `List` with text inputs, e.g., paragraphs

        Returns
        -------
        tokens: `List[List[Token]]`
        """
        texts = [self.text_cleaning(text) for text in document]
        if not self.config.segment_sentences:
            return list(map(self._tokenize, texts[: self.config.max_nr_of_sentences]))
        sentences = [
            sentence.text.strip()
            for doc in self.__nlp__.pipe(texts)
            for sentence in doc.sents
        ]
        return list(map(self._tokenize, sentences[: self.config.max_nr_of_sentences]))

    def tokenize_record(
        self, record: Dict[str, Any], exclude_record_keys: bool
    ) -> List[List[Token]]:
        """Tokenizes a record-like structure containing text inputs

        Use this to keep information about the record-like data structure as input features to the model.

        Parameters
        ----------
        record: `Dict[str, Any]`
            A `Dict` with arbitrary "fields" containing text.
        exclude_record_keys: `bool`
            If enabled, exclude tokens related to record key text

        Returns
        -------
        tokens: `List[List[Token]]`
            A list of tokenized fields as token list
        """
        data = self._sanitize_dict(record)
        if exclude_record_keys:
            return [
                sentence
                for key, value in data.items()
                for sentence in self.tokenize_text(value)
            ]

        return [
            tokenized_key + sentence
            for key, value in data.items()
            for tokenized_key in [self._tokenize(key)]
            for sentence in self.tokenize_text(value)
        ]

    def _value_as_string(self, value: Any) -> str:
        """Converts a value data into its string representation"""
        if isinstance(value, str):
            return value
        if isinstance(value, dict):
            return self._value_as_string(value.values())
        if isinstance(value, Iterable):
            return " ".join(map(self._value_as_string, value))
        return str(value)

    def _sanitize_dict(self, data: Dict[str, Any]) -> Dict[str, str]:
        """Transforms an data dictionary into a dictionary of strings values"""
        return {key: self._value_as_string(value) for key, value in data.items()}

    def _sanitize(self, tokens: List[spacy.tokens.Token]) -> List[Token]:
        """
        Converts spaCy tokens to allennlp tokens. Is a no-op if
        keep_spacy_tokens is True
        """
        if not self.config.use_spacy_tokens:
            tokens = [spacy_to_allennlp_token(token) for token in tokens]
        for start_token in self._start_tokens:
            tokens.insert(0, Token(start_token, 0))
        for end_token in self._end_tokens:
            tokens.append(Token(end_token, -1))
        return tokens


class TransformersTokenizer(Tokenizer):
    """This tokenizer uses the pretrained tokenizers from huggingface's transformers library.

    This means the output will very likely be word pieces depending on the specified pretrained model.

    Parameters
    ----------
    config
        A `TokenizerConfiguration` object
    """

    def __init__(self, config):
        self.pretrained_tokenizer = PretrainedTransformerTokenizer(
            **config.transformers_kwargs
        )
        self._config = config

    def tokenize_document(self, document: List[str]) -> List[List[Token]]:
        return list(map(self._tokenize, document))

    def _tokenize(self, text: str) -> List[Token]:
        return self.pretrained_tokenizer.tokenize(
            text[: self._config.max_sequence_length]
        )

    @property
    def nlp(self) -> Language:
        raise NotImplementedError("For the TransformerTokenizer we have no spaCy nlp")



from allennlp.data import Token as AllennlpToken
from spacy.tokens.token import Token as SpacyToken

from biome.text.configuration import TokenizerConfiguration
from biome.text.tokenizer import Tokenizer

html_text = """
        <!DOCTYPE html>
        <html>
        <body>

        <h1>My First Heading</h1>
        <p>My first paragraph.</p>
        <p>My second paragraph.</p>
        </body>
        </html>
    """


def test_text_cleaning_with_sentence_segmentation():
    tokenizer = Tokenizer(
        TokenizerConfiguration(
            text_cleaning={"rules": ["html_to_text", "strip_spaces"]},
            segment_sentences=True,
        )
    )

    tokenized = tokenizer.tokenize_text(html_text)
    assert len(tokenized) == 2
    assert (
        len(tokenized[0]) == 7
    ), "Expected [My, First, Heading, My, first, paragraph, .]"
    assert len(tokenized[1]) == 4, "Expected [My, second, paragraph, .]"


def test_text_cleaning_with_sentence_segmentation_and_max_sequence():
    tokenizer = Tokenizer(
        TokenizerConfiguration(
            max_sequence_length=8,
            text_cleaning={"rules": ["html_to_text", "strip_spaces"]},
            segment_sentences=True,
        )
    )

    tokenized = tokenizer.tokenize_text(html_text)
    assert len(tokenized) == 2
    assert len(tokenized[0]) == 2, "Expected [My, First]"
    assert len(tokenized[1]) == 2, "Expected [My, second]"


def test_document_cleaning():
    tokenizer = Tokenizer(
        TokenizerConfiguration(
            text_cleaning={"rules": ["html_to_text", "strip_spaces"]},
            segment_sentences=True,
        )
    )

    tokenized = tokenizer.tokenize_document([html_text])
    assert len(tokenized) == 2
    assert (
        len(tokenized[0]) == 7
    ), "Expected [My, First, Heading, My, first, paragraph, .]"
    assert len(tokenized[1]) == 4, "Expected [My, second, paragraph, .]"


def test_using_spacy_tokens():
    tokenizer = Tokenizer(TokenizerConfiguration(use_spacy_tokens=True))
    tokenized = tokenizer.tokenize_text("This is a text")
    assert len(tokenized) == 1
    assert len(tokenized[0]) == 4
    assert all(map(lambda t: isinstance(t, SpacyToken), tokenized[0]))


def test_using_allennlp_tokens():
    tokenizer = Tokenizer(TokenizerConfiguration(use_spacy_tokens=False))
    tokenized = tokenizer.tokenize_text("This is a text")
    assert len(tokenized) == 1
    assert len(tokenized[0]) == 4
    assert all(map(lambda t: isinstance(t, AllennlpToken), tokenized[0]))


def test_set_sentence_segmentation_with_max_number_of_sentences():
    tokenizer = Tokenizer(TokenizerConfiguration(max_nr_of_sentences=2))
    tokenized = tokenizer.tokenize_document(
        [
            "This is a sentence. This is another sentence.",
            "One more sentence here.",
            "Last sentence here.",
        ]
    )
    assert len(tokenized) == 2



import json
import logging
import math
import os
from dataclasses import asdict
from pathlib import Path
from typing import TYPE_CHECKING
from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from typing import Union

import pytorch_lightning as pl
import torch
from allennlp.common import Params
from allennlp.common.util import sanitize
from allennlp.data import Batch
from allennlp.data import Instance
from allennlp.data.data_loaders import TensorDict
from allennlp.data.samplers import BucketBatchSampler
from allennlp.training.optimizers import Optimizer
from pytorch_lightning import Callback
from pytorch_lightning.callbacks import EarlyStopping
from pytorch_lightning.callbacks import LearningRateMonitor
from pytorch_lightning.callbacks import ModelCheckpoint
from pytorch_lightning.loggers import CSVLogger
from pytorch_lightning.loggers import LightningLoggerBase
from pytorch_lightning.loggers import TensorBoardLogger
from pytorch_lightning.loggers import WandbLogger
from pytorch_lightning.utilities.cloud_io import load as pl_load
from torch.utils.data import DataLoader
from torch.utils.data import IterableDataset
from transformers.optimization import get_constant_schedule_with_warmup
from transformers.optimization import get_cosine_schedule_with_warmup
from transformers.optimization import get_linear_schedule_with_warmup

from biome.text import vocabulary
from biome.text.configuration import TrainerConfiguration
from biome.text.configuration import VocabularyConfiguration
from biome.text.dataset import Dataset
from biome.text.dataset import InstanceDataset
from biome.text.errors import EmptyVocabError

if TYPE_CHECKING:
    from biome.text.model import PipelineModel
    from biome.text.pipeline import Pipeline

# We do not require wandb
_HAS_WANDB = False
try:
    import wandb
except ImportError:
    pass
else:
    wandb.ensure_configured()
    _HAS_WANDB = True


_LOGGER = logging.getLogger(__name__)


class Trainer:
    """Class for training and testing a `biome.text.Pipeline`.

    It is basically a light wrapper around the awesome Pytorch Lightning Trainer to define custom defaults and
    facilitate the interaction with our pipelines.

    Parameters
    ----------
    pipeline
        Pipeline to train
    train_dataset
        The training dataset. Default: `None`.
    valid_dataset
        The validation dataset. Default: `None`.
    trainer_config
        The configuration of the trainer. Default: `TrainerConfiguration()`.
    vocab_config
        A `VocabularyConfiguration` to create/extend the pipeline's vocabulary.
        If `"default"` (str), we will use the default configuration `VocabularyConfiguration()`.
        If None, we will leave the pipeline's vocabulary untouched. Default: `"default"`.
    lazy
        If True, instances are lazily loaded from disk, otherwise they are loaded into memory.
        Ignored when passing in `InstanceDataset`s. Default: False.
    """

    def __init__(
        self,
        pipeline: "Pipeline",
        train_dataset: Optional[Union[Dataset, InstanceDataset]] = None,
        valid_dataset: Optional[Union[Dataset, InstanceDataset]] = None,
        trainer_config: Optional[TrainerConfiguration] = None,
        vocab_config: Optional[Union[str, VocabularyConfiguration]] = "default",
        lazy: bool = False,
    ):
        self._pipeline = pipeline
        # since we will make changes to the config, better to make a copy -> asdict returns a deep copy
        self._trainer_config = (
            TrainerConfiguration(**asdict(trainer_config))
            if trainer_config is not None
            else TrainerConfiguration()
        )

        # Use GPU by default if available
        if self._trainer_config.gpus is None and torch.cuda.is_available():
            self._trainer_config.gpus = 1

        self._vocab_config: Optional[VocabularyConfiguration] = (
            VocabularyConfiguration() if vocab_config == "default" else vocab_config
        )
        self._lazy = lazy

        # create instances
        if isinstance(train_dataset, Dataset):
            self._train_instances = train_dataset.to_instances(
                self._pipeline, lazy=self._lazy, tqdm_desc="Loading training instances"
            )
        else:
            self._train_instances = train_dataset

        if isinstance(valid_dataset, Dataset):
            self._valid_instances = valid_dataset.to_instances(
                self._pipeline,
                lazy=self._lazy,
                tqdm_desc="Loading validation instances",
            )
        else:
            self._valid_instances = valid_dataset

        # Maybe we just want to call `self.test`
        if self._train_instances is not None:
            self._setup_for_training()

        self.trainer = pl.Trainer(**self._trainer_config.lightning_params)

    def _setup_for_training(self):
        """Create vocab, configure default loggers/callbacks, create optimizer/lr scheduler, setup best metrics"""
        # create vocab
        if self._vocab_config is not None:
            vocab_datasets = [self._train_instances]
            if (
                self._valid_instances is not None
                and self._vocab_config.include_valid_data
            ):
                vocab_datasets += [self._valid_instances]
            self._pipeline.create_vocab(vocab_datasets, config=self._vocab_config)

        # Check for an empty vocab
        if vocabulary.is_empty(
            self._pipeline.vocab, self._pipeline.config.features.configured_namespaces
        ):
            raise EmptyVocabError(
                "All your features need a non-empty vocabulary for a training!"
            )

        # we give some special attention to these loggers/callbacks
        self._wandb_logger: Optional[WandbLogger] = None
        self._model_checkpoint: Optional[ModelCheckpoint] = None

        # add default callbacks/loggers
        self._trainer_config.callbacks = self._add_default_callbacks()
        if self._trainer_config.logger is not False:
            self._trainer_config.logger = self._add_default_loggers()

        # create optimizer, has to come AFTER creating the vocab!
        self._pipeline.model.optimizer = Optimizer.from_params(
            Params(
                {
                    "model_parameters": self._pipeline.model.named_parameters(),
                    **self._trainer_config.optimizer,
                }
            )
        )

        # create lr scheduler, has to come AFTER creating the optimizer!
        if not (
            self._trainer_config.warmup_steps == 0
            and self._trainer_config.lr_decay is None
        ):
            self._pipeline.model.lr_scheduler = self._create_lr_scheduler()
        else:
            self._pipeline.model.lr_scheduler = None

        # set monitor and mode for best validation metrics
        self._pipeline.model.monitor = self._trainer_config.monitor
        self._pipeline.model.monitor_mode = self._trainer_config.monitor_mode

    def _add_default_loggers(self) -> List[LightningLoggerBase]:
        """Adds optional default loggers and returns the extended list

        Added loggers: CSV, TensorBoard, WandB
        """
        loggers = self._trainer_config.logger
        if loggers is True:
            loggers = []
        elif isinstance(loggers, LightningLoggerBase):
            loggers = [loggers]

        def get_loggers_of_type(logger_type) -> List[LightningLoggerBase]:
            return [logger for logger in loggers if isinstance(logger, logger_type)]

        # csv
        if self._trainer_config.add_csv_logger and not get_loggers_of_type(CSVLogger):
            loggers.append(
                CSVLogger(
                    save_dir=self._trainer_config.default_root_dir or os.getcwd(),
                    name="csv",
                )
            )

        # tensorboard
        if self._trainer_config.add_tensorboard_logger and not get_loggers_of_type(
            TensorBoardLogger
        ):
            loggers.append(
                TensorBoardLogger(
                    save_dir=self._trainer_config.default_root_dir,
                    name="tensorboard",
                )
            )

        # wandb
        if (
            self._trainer_config.add_wandb_logger
            and _HAS_WANDB
            and not get_loggers_of_type(WandbLogger)
        ):
            self._wandb_logger = WandbLogger(
                save_dir=self._trainer_config.default_root_dir,
                project=os.environ.get("WANDB_PROJECT", "biome"),
            )
            loggers.append(self._wandb_logger)
        elif get_loggers_of_type(WandbLogger):
            self._wandb_logger = get_loggers_of_type(WandbLogger)[0]
        # somehow the wandb dir does not get created, i think this is a bug on pl side, have to check it out
        if self._wandb_logger is not None and not os.path.isdir(
            os.path.join(self._wandb_logger.save_dir, "wandb")
        ):
            os.makedirs(os.path.join(self._wandb_logger.save_dir, "wandb"))

        return loggers

    def _add_default_callbacks(self) -> List[Callback]:
        """Adds optional default callbacks and returns the extended list

        Added callbacks: ModelCheckpoint, EarlyStopping, LearningRateMonitor
        """
        callbacks = self._trainer_config.callbacks or []
        if isinstance(callbacks, Callback):
            callbacks = [callbacks]

        def get_callbacks_of_type(callback_type) -> List[Callback]:
            return [
                callback
                for callback in callbacks
                if isinstance(callback, callback_type)
            ]

        # model checkpoint
        if self._trainer_config.checkpoint_callback and not get_callbacks_of_type(
            ModelCheckpoint
        ):
            monitor = self._trainer_config.monitor if self._valid_instances else None
            mode = self._trainer_config.monitor_mode
            save_top_k = (
                self._trainer_config.save_top_k_checkpoints
                if self._valid_instances
                else None
            )
            self._model_checkpoint = ModelCheckpointWithVocab(
                save_top_k=save_top_k, monitor=monitor, mode=mode
            )
            callbacks.append(self._model_checkpoint)
        elif get_callbacks_of_type(ModelCheckpoint):
            self._model_checkpoint = get_callbacks_of_type(ModelCheckpoint)[0]

        # early stopping
        if (
            self._trainer_config.add_early_stopping
            and self._valid_instances is not None
            and not get_callbacks_of_type(EarlyStopping)
        ):
            callbacks.append(
                EarlyStopping(
                    monitor=self._trainer_config.monitor,
                    mode=self._trainer_config.monitor_mode,
                    patience=self._trainer_config.patience,
                )
            )

        # lr monitor
        if self._trainer_config.add_lr_monitor is None and (
            self._trainer_config.warmup_steps != 0
            or self._trainer_config.lr_decay is not None
        ):
            self._trainer_config.add_lr_monitor = True
        if self._trainer_config.add_lr_monitor and not get_callbacks_of_type(
            LearningRateMonitor
        ):
            callbacks.append(LearningRateMonitor(logging_interval="step"))

        return callbacks

    def _create_lr_scheduler(self) -> Dict:
        """Returns one of three default schedulers

        Possibilities: constant/linear/cosine schedule with or without warmup
        """
        steps_per_epoch = math.ceil(
            len(self._train_instances) / self._trainer_config.batch_size
        )
        try:
            training_steps = min(
                self._trainer_config.max_steps,
                self._trainer_config.max_epochs * steps_per_epoch,
            )
        # One or both of the max_* is None:
        except TypeError:
            training_steps = (
                self._trainer_config.max_steps
                # 1000 is the default of the lightning trainer
                or (self._trainer_config.max_epochs or 1000) * steps_per_epoch
            )

        if self._trainer_config.lr_decay == "linear":
            scheduler = get_linear_schedule_with_warmup(
                optimizer=self._pipeline.model.optimizer,
                num_warmup_steps=self._trainer_config.warmup_steps,
                num_training_steps=training_steps,
            )
        elif self._trainer_config.lr_decay == "cosine":
            scheduler = get_cosine_schedule_with_warmup(
                optimizer=self._pipeline.model.optimizer,
                num_warmup_steps=self._trainer_config.warmup_steps,
                num_training_steps=training_steps,
            )
        else:
            scheduler = get_constant_schedule_with_warmup(
                optimizer=self._pipeline.model.optimizer,
                num_warmup_steps=self._trainer_config.warmup_steps,
            )

        return {
            "scheduler": scheduler,
            "interval": "step",
            "name": "learning_rate",
        }

    def fit(
        self, output_dir: Optional[Union[str, Path]] = "output", exist_ok: bool = False
    ):
        """Train the pipeline

        At the end of the training the pipeline will load the weights from the best checkpoint.

        Parameters
        ----------
        output_dir
            If specified, save the trained pipeline to this directory. Default: 'output'.
        exist_ok
            If True, overwrite the content of `output_dir`. Default: False.
        """
        if self._train_instances is None:
            _LOGGER.error(
                "You need training data to fit your model, please provide it on `Trainer.__init__`."
            )
            return

        try:
            output_dir = Path(output_dir)
        except TypeError:
            output_dir = None
        else:
            output_dir.mkdir(exist_ok=exist_ok)

        # create dataloaders
        train_dataloader = create_dataloader(
            self._train_instances,
            batch_size=self._trainer_config.batch_size,
            num_workers=self._trainer_config.num_workers_for_dataloader,
        )
        valid_dataloader = (
            create_dataloader(
                self._valid_instances,
                batch_size=self._trainer_config.batch_size,
                num_workers=self._trainer_config.num_workers_for_dataloader,
            )
            if self._valid_instances is not None
            else None
        )

        # log config to wandb
        if self._wandb_logger is not None:
            config = {
                "pipeline": self._pipeline.config.as_dict(),
                "trainer": self._trainer_config.as_dict(),
            }
            self._wandb_logger.experiment.config.update(config)

        # training
        try:
            self.trainer.fit(
                self._pipeline.model,
                train_dataloader=train_dataloader,
                val_dataloaders=valid_dataloader,
            )
        finally:
            if self._model_checkpoint:
                self._load_best_weights()
            if output_dir:
                self._pipeline.save(output_dir)
                with (output_dir / "metrics.json").open("w") as file:
                    json.dump(sanitize(self.trainer.logged_metrics), file)
                self._trainer_config.to_yaml(output_dir / "trainer_config.yaml")
            if self._wandb_logger is not None:
                self._wandb_logger.experiment.finish()

    def _load_best_weights(self):
        """Load weights from the best model checkpoint"""
        checkpoint_path = self._model_checkpoint.best_model_path
        if checkpoint_path:
            _LOGGER.debug("Loading best weights")
            checkpoint = pl_load(
                checkpoint_path, map_location=lambda storage, loc: storage
            )
            self._pipeline.model.load_state_dict(checkpoint["state_dict"])

    def test(
        self,
        test_dataset: Union[Dataset, InstanceDataset],
        batch_size: Optional[int] = None,
        output_dir: Optional[Union[str, Path]] = None,
        verbose: bool = True,
    ) -> Dict[str, Any]:
        """Evaluate your model on a test dataset

        Parameters
        ----------
        test_dataset
            The test data set.
        batch_size
            The batch size. If None (default), we will use the batch size specified in the `TrainerConfiguration`.
        output_dir
            Save a `metrics.json` to this output directory. Default: None.
        verbose
            If True, prints the test results. Default: True.

        Returns
        -------
        Dict[str, Any]
            A dictionary with the metrics
        """
        # load instances
        if isinstance(test_dataset, Dataset):
            self._train_instances = test_dataset.to_instances(
                self._pipeline, lazy=self._lazy, tqdm_desc="Loading test instances"
            )
        else:
            self._train_instances = test_dataset

        # create dataloader
        test_dataloader = create_dataloader(
            self._train_instances,
            batch_size=batch_size or self._trainer_config.batch_size,
            num_workers=self._trainer_config.num_workers_for_dataloader,
        )

        metrics = self.trainer.test(
            self._pipeline.model, test_dataloaders=test_dataloader, verbose=verbose
        )[0]

        if output_dir:
            with (output_dir / "metrics.json").open("w") as file:
                json.dump(sanitize(metrics), file)

        return metrics


class ModelCheckpointWithVocab(ModelCheckpoint):
    def on_pretrain_routine_start(self, trainer, pl_module: "PipelineModel"):
        super().on_pretrain_routine_start(trainer, pl_module)
        if os.path.isdir(self.dirpath):
            pl_module.vocab.save_to_files(os.path.join(self.dirpath, "vocabulary"))


def create_dataloader(
    instance_dataset: InstanceDataset,
    batch_size: int = 16,
    num_workers: int = 0,
) -> DataLoader:
    """Returns a pytorch DataLoader for AllenNLP instances

    Parameters
    ----------
    instance_dataset
        The dataset of instances for the DataLoader
    batch_size
        Batch size
    num_workers
        How many subprocesses to use for data loading. 0 means that the data will be loaded in the main process.
        Default: 0

    Returns
    -------
    data_loader
    """
    return DataLoader(
        instance_dataset,
        batch_size=batch_size,
        num_workers=num_workers,
        collate_fn=allennlp_collate,
    )


def allennlp_collate(instances: List[Instance]) -> TensorDict:
    batch = Batch(instances)
    return batch.as_tensor_dict(batch.get_padding_lengths())



from typing import Dict

import pandas as pd
import pytest
from allennlp.data import Batch

from biome.text import Dataset
from biome.text import Pipeline
from biome.text import Trainer
from biome.text import TrainerConfiguration
from biome.text import vocabulary
from biome.text.modules.heads.task_prediction import Entity
from biome.text.modules.heads.task_prediction import TokenClassificationPrediction


@pytest.fixture
def training_dataset() -> Dataset:
    df = pd.DataFrame(
        {
            "text": [
                "This is a simple NER test",
                "This is a simple NER test with misaligned spans",
                "No NER here",
            ],
            "entities": [
                [{"start": 17, "end": 20, "label": "NER"}],
                [{"start": 17, "end": 22, "label": "NER"}],
                [],
            ],
        }
    )

    return Dataset.from_pandas(df)


@pytest.fixture
def pipeline_dict() -> Dict:
    pipeline_dict = {
        "name": "biome-bimpm",
        "features": {"word": {"embedding_dim": 2}},
        "head": {
            "type": "TokenClassification",
            "labels": ["NER"],
            "label_encoding": "BIOUL",
        },
    }

    return pipeline_dict


@pytest.fixture
def trainer_config() -> TrainerConfiguration:
    return TrainerConfiguration(
        max_epochs=1, batch_size=1, optimizer={"type": "adam", "lr": 0.01}, gpus=0
    )


def test_tokenization_with_blank_tokens(pipeline_dict):
    pipeline = Pipeline.from_config(pipeline_dict)
    predictions = pipeline.predict(text="Test this text \n \n", entities=[])
    assert len(predictions["tags"][0]) == 4


def test_train(pipeline_dict, training_dataset, trainer_config, tmp_path):
    pipeline = Pipeline.from_config(pipeline_dict)

    assert pipeline.output == ["entities", "tags"]

    assert pipeline.head.span_labels == ["NER"]
    assert pipeline.head.labels == ["B-NER", "I-NER", "U-NER", "L-NER", "O"]

    trainer = Trainer(
        pipeline=pipeline, train_dataset=training_dataset, trainer_config=trainer_config
    )
    trainer.fit(tmp_path / "ner_experiment")


class TestMakeTaskPrediction:
    def test_pretokenized_input(self, pipeline_dict):
        pipeline = Pipeline.from_config(pipeline_dict)
        output = self._input_top_k2(pipeline)
        expected_output = TokenClassificationPrediction(
            tags=[["O", "O", "O", "U-NER"], ["O", "B-NER", "I-NER", "L-NER"]],
            entities=[
                [Entity(start_token=3, end_token=4, label="NER")],
                [Entity(start_token=1, end_token=4, label="NER")],
            ],
            scores=[2, 1],
        )

        assert output == expected_output

    def test_untokenized_input(self, pipeline_dict):
        pipeline = Pipeline.from_config(pipeline_dict)
        output = self._input_top_k2(pipeline, pretokenized=False)
        expected_output = TokenClassificationPrediction(
            tags=[["O", "O", "O", "U-NER"], ["O", "B-NER", "I-NER", "L-NER"]],
            entities=[
                [Entity(start_token=3, end_token=4, label="NER", start=10, end=14)],
                [Entity(start_token=1, end_token=4, label="NER", start=5, end=14)],
            ],
            scores=[2, 1],
        )

        assert output == expected_output

    @staticmethod
    def _input_top_k2(pipeline, pretokenized=True):
        raw_text = ["this", "is", "a", "test"] if pretokenized else "this is a test"
        tag_idx_sequence = [
            vocabulary.index_for_label(pipeline.backbone.vocab, tag)
            for tag in ["O", "O", "O", "U-NER"]
        ]
        tag_idx_sequence2 = [
            vocabulary.index_for_label(pipeline.backbone.vocab, tag)
            for tag in ["O", "B-NER", "I-NER", "L-NER"]
        ]
        viterbi_paths = [(tag_idx_sequence, 2), (tag_idx_sequence2, 1)]
        single_forward_output = dict(
            viterbi_paths=viterbi_paths,
            raw_text=raw_text,
        )

        return pipeline.head._make_task_prediction(single_forward_output, None)


def test_preserve_pretokenization(pipeline_dict):
    pipeline = Pipeline.from_config(pipeline_dict)
    tokens = ["test", "this", "pre tokenized", "text"]
    prediction = pipeline.predict(tokens)
    assert len(prediction["tags"][0]) == len(tokens)


def test_metrics(pipeline_dict):
    pipeline = Pipeline.from_config(pipeline_dict)
    instance = pipeline.head.featurize(text="test this".split(), tags=["U-NER", "O"])
    batch = Batch([instance])
    batch.index_instances(pipeline.vocab)

    pipeline.head.forward(**batch.as_tensor_dict())
    # validation metric should have never been called
    assert pipeline.head._metrics.get_dict()["accuracy"].total_count == 2
    assert pipeline.head._metrics.get_dict(is_train=False)["accuracy"].total_count == 0

    train_metrics = pipeline.head.get_metrics(reset=True)
    expected_metric_names = ["accuracy"] + [
        f"{metric}-{label}"
        for metric in ["precision", "recall", "f1-measure"]
        for label in ["NER", "overall"]
    ]
    print(train_metrics)
    assert all(name in train_metrics for name in expected_metric_names)

    pipeline.head.training = False
    pipeline.head.forward(**batch.as_tensor_dict())
    # training metric should have never been called after its reset
    assert pipeline.head._metrics.get_dict()["accuracy"].total_count == 0
    assert pipeline.head._metrics.get_dict(is_train=False)["accuracy"].total_count == 2

    valid_metrics = pipeline.head.get_metrics()
    assert all(name in valid_metrics for name in expected_metric_names)



import copy
from typing import Any
from typing import Dict

from allennlp.common import Params
from allennlp.training.metrics import Metric


class Metrics:
    """Stores two dictionaries of identical metrics, one for training and one for validation.

    Parameters
    ----------
    **kwargs
        The key defines the name of the metric, the value must be a dictionary that can be used to instantiate a
        child class of `allennlp.training.metrics.Metric` via its `from_params` method.

    Examples
    --------
    >>> from allennlp.training.metrics import Metric
    >>> metrics = Metrics(accuracy={"type": "categorical_accuracy"}, f1={"type": "fbeta"})
    >>> for metric in metrics.get_dict(is_train=False).values():
    ...     assert isinstance(metric, Metric)
    """

    def __init__(self, **kwargs: Dict[str, Any]):
        self.training_metrics = {}
        self.validation_metrics = {}
        for name, metric_kwargs in kwargs.items():
            # We need a special logic for the vocabulary, we do not want to deep copy it,
            # and it cannot be used in Params
            vocab = metric_kwargs.pop("vocabulary", None)
            self.training_metrics[name] = Metric.from_params(
                Params(copy.deepcopy(metric_kwargs)),
                **{} if vocab is None else {"vocabulary": vocab}
            )
            self.validation_metrics[name] = Metric.from_params(
                Params(metric_kwargs), **{} if vocab is None else {"vocabulary": vocab}
            )

    def get_dict(self, is_train: bool = True) -> Dict[str, Metric]:
        if is_train:
            return self.validation_metrics
        return self.training_metrics



from typing import Dict

import pytest

from biome.text import Dataset
from biome.text import Pipeline
from biome.text import Trainer
from biome.text import TrainerConfiguration


@pytest.fixture
def training_dataset() -> Dataset:
    """Creating the dataframe."""
    data = {
        "text": [
            "this is a text",
            "my name is dani",
            "this is a table",
            "my name is paco",
        ],
    }
    return Dataset.from_dict(data)


@pytest.fixture
def pipeline_dict() -> Dict:
    """Creating the pipeline dictionary"""

    pipeline_dict = {
        "name": "lm",
        "features": {
            "word": {"embedding_dim": 50, "lowercase_tokens": True, "trainable": True},
            "char": {
                "embedding_dim": 50,
                "dropout": 0.1,
                "encoder": {
                    "type": "gru",
                    "hidden_size": 10,
                    "num_layers": 1,
                    "bidirectional": True,
                },
            },
        },
        "encoder": {
            "type": "gru",
            "num_layers": 1,
            "hidden_size": 10,
            "bidirectional": True,
        },
        "head": {"type": "LanguageModelling", "dropout": 0.1, "bidirectional": True},
    }

    return pipeline_dict


@pytest.fixture
def trainer_config() -> TrainerConfiguration:
    return TrainerConfiguration(
        max_epochs=2,
        optimizer={"type": "adam", "amsgrad": True, "lr": 0.002},
        gpus=0,
    )


def test_train(pipeline_dict, training_dataset, trainer_config, tmp_path):
    """Testing the correct working of prediction, vocab creating and training"""

    pipeline = Pipeline.from_config(pipeline_dict)
    pipeline.predict(text="my name is juan")

    trainer = Trainer(
        pipeline=pipeline,
        train_dataset=training_dataset,
        valid_dataset=training_dataset,
        trainer_config=trainer_config,
    )
    trainer.fit(tmp_path / "lm")



from allennlp.modules import Seq2SeqEncoder
from allennlp.modules import TimeDistributed


class TimeDistributedEncoder(Seq2SeqEncoder):
    """Wraps a Seq2SeqEncoder into a TimeDistributed module and implements the Seq2SeqEncoder API"""

    def __init__(self, encoder: Seq2SeqEncoder):
        super(TimeDistributedEncoder, self).__init__()

        self._input_dim = encoder.get_input_dim()
        self._output_dim = encoder.get_output_dim()
        self._is_bidirectional = (
            hasattr(encoder, "is_bidirectional") and encoder.is_bidirectional()
        )

        self._encoder = TimeDistributed(encoder)

    def forward(self, *input, **inputs):
        return self._encoder(*input, **inputs)

    def is_bidirectional(self) -> bool:
        return self._is_bidirectional

    def get_output_dim(self) -> int:
        return self._output_dim

    def get_input_dim(self):
        return self._input_dim



from typing import Dict
from typing import List
from typing import Optional
from typing import Union

import numpy
from allennlp.data import Instance

from biome.text.backbone import ModelBackbone
from biome.text.modules.configuration import ComponentConfiguration
from biome.text.modules.configuration import FeedForwardConfiguration
from biome.text.modules.configuration import Seq2SeqEncoderConfiguration
from biome.text.modules.configuration import Seq2VecEncoderConfiguration
from biome.text.modules.heads import DocumentClassification
from biome.text.modules.heads.task_prediction import RecordClassificationPrediction


class RecordClassification(DocumentClassification):
    """
    Task head for data record  classification.
    Accepts a variable data inputs and apply featuring over defined record keys.

    This head applies a doc2vec architecture from a structured record data input

    Parameters
    ----------
    backbone
        The backbone of your model. Must not be provided when initiating with `Pipeline.from_config`.
    labels
        A list of labels for your classification task.
    token_pooler
        The pooler at token level to provide one vector per record field. Default: `BagOfEmbeddingsEncoder`.
    fields_encoder
        An optional sequence to sequence encoder that contextualizes the record field representations. Default: None.
    fields_pooler
        The pooler at sentence level to provide a vector for the whole record. Default: `BagOfEmbeddingsEncoder`.
    feedforward
        An optional feedforward layer applied to the output of the fields pooler. Default: None.
    multilabel
        Is this a multi label classification task? Default: False
    label_weights
        A list of weights for each label. The weights must be in the same order as the `labels`.
        You can also provide a dictionary that maps the label to its weight. Default: None.
    """

    def __init__(
        self,
        backbone: ModelBackbone,
        labels: List[str],
        record_keys: List[str],
        token_pooler: Optional[Seq2VecEncoderConfiguration] = None,
        fields_encoder: Optional[Seq2SeqEncoderConfiguration] = None,
        fields_pooler: Optional[Seq2VecEncoderConfiguration] = None,
        feedforward: Optional[FeedForwardConfiguration] = None,
        multilabel: Optional[bool] = False,
        label_weights: Optional[Union[List[float], Dict[str, float]]] = None,
    ) -> None:

        super().__init__(
            backbone,
            labels=labels,
            token_pooler=token_pooler,
            sentence_encoder=fields_encoder,
            sentence_pooler=fields_pooler,
            feedforward=feedforward,
            multilabel=multilabel,
            label_weights=label_weights,
        )

        self._inputs = record_keys

    def inputs(self) -> Optional[List[str]]:
        """The inputs names are determined by configured record keys"""
        return self._inputs

    def featurize(
        self, label: Optional[Union[str, List[str]]] = None, **inputs
    ) -> Instance:

        record = {input_key: inputs[input_key] for input_key in self._inputs}
        instance = self.backbone.featurizer(record, to_field=self.forward_arg_name)

        return self._add_label(instance, label)

    def _make_task_prediction(
        self,
        single_forward_output: Dict[str, numpy.ndarray],
        instance: Instance,
    ) -> RecordClassificationPrediction:
        labels, probabilities = self._compute_labels_and_probabilities(
            single_forward_output
        )

        return RecordClassificationPrediction(
            labels=labels, probabilities=probabilities
        )


class RecordClassificationConfiguration(ComponentConfiguration[RecordClassification]):
    """Lazy initialization for document classification head components"""

    pass



import logging
from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from typing import Union
from typing import cast

import numpy
import torch
from allennlp.data import Instance
from allennlp.data import TextFieldTensors
from allennlp.data.fields import SequenceLabelField
from allennlp.data.fields import TextField
from allennlp.modules.seq2vec_encoders import BagOfEmbeddingsEncoder
from allennlp.nn.util import get_text_field_mask

from biome.text.backbone import ModelBackbone
from biome.text.featurizer import FeaturizeError
from biome.text.helpers import tags_from_offsets
from biome.text.modules.configuration import ComponentConfiguration
from biome.text.modules.configuration import EmbeddingConfiguration
from biome.text.modules.configuration import FeedForwardConfiguration
from biome.text.modules.configuration import Seq2VecEncoderConfiguration
from biome.text.modules.heads.classification.classification import ClassificationHead

# from biome.text.modules.encoders.multi_head_self_attention_encoder import MultiheadSelfAttentionEncoder
from biome.text.modules.heads.task_prediction import RelationClassificationPrediction


class RelationClassification(ClassificationHead):
    """
    Task head for relation classification

    Parameters
    ----------
    backbone
        The backbone of your model. Must not be provided when initiating with `Pipeline.from_config`.
    labels
        A list of labels for your classification task.
    entities_embedder
        The embedder for the entity tags.
    entity_encoding
        The encoding scheme of the entity tags. Default: BIOUL.
    pooler
        The pooler of the output sequence from the backbone model. Default: `BagOfEmbeddingsEncoder`.
    feedforward
        An optional feedforward layer applied to the output of the pooler. Default: None.
    multilabel
        Is this a multi label classification task? Default: False
    label_weights
        A list of weights for each label. The weights must be in the same order as the `labels`.
        You can also provide a dictionary that maps the label to its weight. Default: None.
    """

    _TEXT_ARG_NAME_IN_FORWARD = "text"
    _LABEL_ARG_NAME_IN_FORWARD = "label"
    _LOGGER = logging.getLogger(__name__)

    def __init__(
        self,
        backbone: ModelBackbone,
        labels: List[str],
        entities_embedder: EmbeddingConfiguration,
        entity_encoding: Optional[str] = "BIOUL",
        pooler: Optional[Seq2VecEncoderConfiguration] = None,
        feedforward: Optional[FeedForwardConfiguration] = None,
        multilabel: bool = False,
        label_weights: Optional[Union[List[float], Dict[str, float]]] = None,
        # self_attention: Optional[MultiheadSelfAttentionEncoder] = None
    ) -> None:

        super().__init__(
            backbone=backbone,
            labels=labels,
            multilabel=multilabel,
            label_weights=label_weights,
        )

        self._label_encoding = entity_encoding
        self._entity_tags_namespace = "entities"

        self.entities_embedder = entities_embedder.compile()

        encoding_output_dim = (
            self.backbone.encoder.get_output_dim()
            + self.entities_embedder.get_output_dim()
        )
        self.pooler = (
            pooler.input_dim(encoding_output_dim).compile()
            if pooler
            else BagOfEmbeddingsEncoder(encoding_output_dim)
        )

        self.feedforward = (
            None
            if not feedforward
            else feedforward.input_dim(self.pooler.get_output_dim()).compile()
        )

        self._classification_layer = torch.nn.Linear(
            (self.feedforward or self.pooler).get_output_dim(), self.num_labels
        )
        # self.self_attention = self_attention

    def featurize(
        self,
        text: Union[str, List[str], Dict[str, str]],
        entities: List[Dict],
        label: Optional[Union[str, List[str]]] = None,
    ) -> Instance:
        instance = self.backbone.featurizer(
            text,
            to_field=self._TEXT_ARG_NAME_IN_FORWARD,
            aggregate=True,
            exclude_record_keys=True,
        )

        doc = self.backbone.tokenizer.nlp(text)
        entity_tags = tags_from_offsets(doc, entities, self._label_encoding)

        if "-" in entity_tags:
            raise FeaturizeError(
                f"Could not align spans with tokens for following example: '{text}' {entities}"
            )

        try:
            instance.add_field(
                "entities",
                SequenceLabelField(
                    entity_tags,
                    sequence_field=cast(TextField, instance["text"]),
                    label_namespace=self._entity_tags_namespace,
                ),
            )
        except Exception as error:
            raise FeaturizeError(
                f"Could not create SequenceLabelField for {(text, entity_tags)}"
            ) from error

        return self._add_label(
            instance, label, to_field=self._LABEL_ARG_NAME_IN_FORWARD
        )

    def forward(  # type: ignore
        self,
        text: TextFieldTensors,
        entities: torch.IntTensor,
        label: torch.IntTensor = None,
    ) -> Dict[str, Any]:

        mask = get_text_field_mask(text)
        embedded_text = self.backbone.forward(text, mask)

        embedded_ents = self.entities_embedder(entities)
        embedded_text = torch.cat((embedded_text, embedded_ents), dim=-1)
        """
        if self.self_attention:
            embedded_text = self.self_attention(embedded_text)
        """
        embedded_text = self.pooler(embedded_text, mask=mask)

        if self.feedforward is not None:
            embedded_text = self.feedforward(embedded_text)

        logits = self._classification_layer(embedded_text)

        return self._make_forward_output(logits=logits, label=label)

    def _make_task_prediction(
        self,
        single_forward_output: Dict[str, numpy.ndarray],
        instance: Instance,
    ) -> RelationClassificationPrediction:
        labels, probabilities = self._compute_labels_and_probabilities(
            single_forward_output
        )

        return RelationClassificationPrediction(
            labels=labels, probabilities=probabilities
        )


class RelationClassificationConfiguration(
    ComponentConfiguration[RelationClassification]
):
    """Configuration for classification head components"""

    pass



import os
import tempfile

import pytest
from pytorch_lightning.loggers import CSVLogger
from pytorch_lightning.loggers import LoggerCollection
from pytorch_lightning.loggers import MLFlowLogger
from pytorch_lightning.loggers import TensorBoardLogger
from pytorch_lightning.loggers import WandbLogger

from biome.text import Dataset
from biome.text import Pipeline
from biome.text import Trainer
from biome.text.configuration import TrainerConfiguration
from biome.text.trainer import create_dataloader


@pytest.fixture
def dataset(resources_data_path) -> Dataset:
    return Dataset.from_csv(
        paths=str(resources_data_path / "business.cat.2k.valid.csv")
    )


@pytest.fixture
def pipeline_dict() -> dict:
    return {
        "name": "german_business_names",
        "features": {
            "word": {"embedding_dim": 16, "lowercase_tokens": True},
        },
        "head": {
            "type": "TextClassification",
            "labels": [
                "Unternehmensberatungen",
                "Friseure",
                "Tiefbau",
                "Dienstleistungen",
                "Gebrauchtwagen",
                "Restaurants",
                "Architekturbüros",
                "Elektriker",
                "Vereine",
                "Versicherungsvermittler",
                "Sanitärinstallationen",
                "Edv",
                "Maler",
                "Physiotherapie",
                "Werbeagenturen",
                "Apotheken",
                "Vermittlungen",
                "Hotels",
                "Autowerkstätten",
                "Elektrotechnik",
                "Allgemeinärzte",
                "Handelsvermittler Und -vertreter",
            ],
        },
    }


def test_default_root_dir(change_to_tmp_working_dir, pipeline_dict, dataset):
    pl = Pipeline.from_config(pipeline_dict)
    trainer = Trainer(pl, train_dataset=dataset)
    assert trainer.trainer.default_root_dir == str(
        change_to_tmp_working_dir / "training_logs"
    )


def test_deep_copy_of_trainer_config(pipeline_dict, dataset):
    pl = Pipeline.from_config(pipeline_dict)
    trainer_config = TrainerConfiguration()
    trainer = Trainer(pl, train_dataset=dataset, trainer_config=trainer_config)
    assert trainer_config is not trainer._trainer_config


@pytest.mark.parametrize(
    "input_kwargs,expected_loggers",
    [
        ({}, ["csv", "tensorboard", "wandb"]),
        ({"logger": False}, []),
        (
            {
                "logger": MLFlowLogger(
                    tracking_uri=os.path.join(tempfile.gettempdir(), "mlruns")
                ),
                "add_wandb_logger": False,
            },
            ["csv", "tensorboard", "mlflow"],
        ),
        (
            {
                "logger": [
                    MLFlowLogger(
                        tracking_uri=os.path.join(tempfile.gettempdir(), "mlruns")
                    ),
                    CSVLogger(save_dir=tempfile.gettempdir()),
                ],
                "add_wandb_logger": False,
                "add_tensorboard_logger": False,
            },
            ["csv", "mlflow"],
        ),
    ],
)
def test_add_default_loggers(
    input_kwargs, expected_loggers, pipeline_dict, dataset, tmp_path
):
    trainer_config = TrainerConfiguration(
        **input_kwargs, default_root_dir=str(tmp_path)
    )
    trainer = Trainer(
        Pipeline.from_config(pipeline_dict),
        train_dataset=dataset,
        trainer_config=trainer_config,
    )
    if input_kwargs.get("logger") is not False:
        assert isinstance(trainer.trainer.logger, LoggerCollection)
        assert len(trainer.trainer.logger.experiment) == len(expected_loggers)
    else:
        assert trainer._trainer_config.logger is False

    def loggers_include(logger_type) -> bool:
        return any(
            [
                isinstance(logger, logger_type)
                for logger in trainer._trainer_config.logger
            ]
        )

    for logger in expected_loggers:
        if logger == "csv":
            assert loggers_include(CSVLogger)
        if logger == "tensorboard":
            assert loggers_include(TensorBoardLogger)
        if logger == "wandb":
            assert loggers_include(WandbLogger)
            assert (tmp_path / "wandb").is_dir()
        if logger == "mlflow":
            assert loggers_include(MLFlowLogger)


def test_pipeline_test(pipeline_dict, dataset, tmp_path):
    import json

    pl = Pipeline.from_config(pipeline_dict)
    trainer = Trainer(pl)
    first_metrics = trainer.test(dataset, output_dir=tmp_path, batch_size=16)
    assert "test_loss" in first_metrics

    assert (tmp_path / "metrics.json").is_file()
    with (tmp_path / "metrics.json").open() as file:
        assert "test_loss" in json.load(file)

    assert pl.evaluate(dataset)["test_loss"] == pytest.approx(
        first_metrics["test_loss"]
    )



from allennlp.modules import BiMpmMatching
from allennlp.modules import Embedding
from allennlp.modules import FeedForward
from allennlp.modules import Seq2SeqEncoder
from allennlp.modules import Seq2VecEncoder

from .defs import ComponentConfiguration


class Seq2VecEncoderConfiguration(ComponentConfiguration[Seq2VecEncoder]):
    """Layer spec for Seq2VecEncoder components"""

    pass


class Seq2SeqEncoderConfiguration(ComponentConfiguration[Seq2SeqEncoder]):
    """Layer spec for Seq2SeqEncoder components"""

    pass


class FeedForwardConfiguration(ComponentConfiguration[FeedForward]):
    """Layer spec for FeedForward components"""

    pass


class BiMpmMatchingConfiguration(ComponentConfiguration[BiMpmMatching]):
    """Layer spec for BiMpmMatching components"""

    pass


class EmbeddingConfiguration(ComponentConfiguration[Embedding]):
    """Layer spec for Embedding components"""

    pass



import inspect
from typing import List

import click
import uvicorn
from allennlp.common.util import sanitize
from click import Path
from fastapi import FastAPI
from fastapi import HTTPException
from fastapi.exceptions import RequestValidationError
from fastapi.responses import PlainTextResponse
from pydantic import BaseConfig
from pydantic import create_model
from starlette.exceptions import HTTPException as StarletteHTTPException

from biome.text import Pipeline


@click.command()
@click.argument("pipeline_path", type=Path(exists=True))
@click.option(
    "--port",
    "-p",
    type=int,
    default=9999,
    show_default=True,
    help="Port on which to serve the REST API.",
)
@click.option(
    "--predictions_dir",
    "-pd",
    type=click.Path(),
    default=None,
    help="Path to log raw predictions from the service.",
)
@click.option(
    "--host",
    type=str,
    default="0.0.0.0",
    help="Host of the underlying uvicorn server",
)
def serve(pipeline_path: str, port: int, predictions_dir: str, host: str) -> None:
    """Serves the pipeline predictions as a REST API

    PIPELINE_PATH is the path to a pretrained pipeline (model.tar.gz file).
    """
    pipeline = Pipeline.from_pretrained(pipeline_path)
    pipeline._model.eval()

    if predictions_dir:
        pipeline.init_prediction_logger(predictions_dir)

    return _serve(pipeline, port, host)


def _serve(pipeline: Pipeline, port: int = 9999, host: str = "0.0.0.0"):
    """Serves an pipeline as rest api"""
    predict_parameters = inspect.signature(pipeline.predict).parameters
    model_parameters = {
        name: (
            par.annotation,
            None,  # We need a default value to allow for batch predictions!
        )
        for name, par in predict_parameters.items()
        if par.default == inspect.Parameter.empty
    }
    optional_parameters = {
        name: (par.annotation, par.default)
        for name, par in predict_parameters.items()
        # The batch parameter needs an extra logic to allow for a proper BaseModel for it
        if par.default != inspect.Parameter.empty and name != "batch"
    }

    class Config(BaseConfig):
        extra = "forbid"

    ModelInput = create_model("ModelInput", **model_parameters, __config__=Config)
    PredictInput = create_model(
        "PredictInput",
        **model_parameters,
        batch=(List[ModelInput], None),
        **optional_parameters,
        __config__=Config,
    )

    class http_error_handling:
        """Error handling for http error transcription"""

        def __enter__(self):
            pass

        def __exit__(self, exc_type, exc_val, exc_tb):
            if isinstance(exc_val, Exception):
                # Common http error handling
                raise HTTPException(status_code=500, detail=str(exc_val))

    def make_app() -> FastAPI:
        app = FastAPI()

        error_msg = f"\nCheck the docs at '0.0.0.0:{port}/docs' for an example of a valid request body."

        @app.exception_handler(RequestValidationError)
        async def validation_exception_handler(request, exc):
            return PlainTextResponse(str(exc) + error_msg, status_code=400)

        @app.exception_handler(StarletteHTTPException)
        async def http_exception_handler(request, exc):
            if exc.status_code == 400:
                return PlainTextResponse(
                    str(exc.detail) + error_msg, status_code=exc.status_code
                )
            else:
                return PlainTextResponse(str(exc.detail), status_code=exc.status_code)

        @app.post("/predict", tags=["Pipeline"])
        async def predict(predict_input: PredictInput):
            """Returns a prediction given some input data

            Parameters
            ----------
            - **args/kwargs:** See the Example Value for the Request body below.
            If provided, the **batch** parameter will be ignored.
            - **batch:** A list of dictionaries that represents a batch of inputs.
            The dictionary keys must comply with the **args/kwargs**.
            Predicting batches should typically be faster than repeated calls with **args/kwargs**.
            - **add_tokens:** If true, adds a 'tokens' key in the prediction that contains the tokenized input.
            - **add_attributions:** If true, adds a 'attributions' key that contains attributions of the input to the prediction.
            - **attributions_kwargs:** This dict is directly passed on to the `TaskHead.compute_attributions()`.

            Returns
            -------
            - **predictions:** A dictionary or a list of dictionaries containing the predictions and additional information.
            """
            with http_error_handling():
                return sanitize(
                    pipeline.predict(**predict_input.dict(skip_defaults=True))
                )

        @app.get("/config", tags=["Pipeline"])
        async def config():
            """The configuration of the pipeline"""
            with http_error_handling():
                return pipeline.config.as_dict()

        @app.get("/_status", tags=["REST service"])
        async def status():
            with http_error_handling():
                return {"ok": True}

        return app

    uvicorn.run(make_app(), host=host, port=port)



from typing import Dict

import pytest

from biome.text import Dataset
from biome.text import Pipeline
from biome.text import Trainer
from biome.text import TrainerConfiguration
from biome.text.modules.heads.task_prediction import Attribution


@pytest.fixture
def training_dataset() -> Dataset:
    """Creating the dataframe."""
    data = {
        "record1": [
            {"@first_name": "Hans", "@last_name": "Peter"},
            {"@first_name": "Heinrich", "@last_name": "Meier"},
            {"@first_name": "Hans", "@last_name": "Peter"},
        ],
        "record2": [
            {"@first_name": "Hans", "@last_name": "Petre"},
            {"@first_name": "Heinz", "@last_name": "Meier"},
            {"@first_name": "Hansel", "@last_name": "Peter"},
        ],
        "label": ["duplicate", "not_duplicate", "duplicate"],
    }

    return Dataset.from_dict(data)


@pytest.fixture
def pipeline_dict() -> Dict:
    """Creating the pipeline dictionary"""

    pipeline_dict = {
        "name": "biome-bimpm",
        "tokenizer": {"text_cleaning": {"rules": ["strip_spaces"]}},
        "features": {
            # "tokens": {
            #     "indexer": {"type": "single_id", "lowercase_tokens": True},
            #     "embedder": {
            #         "type": "embedding",
            #         "embedding_dim": 32,
            #         "padding_index": 0,
            #     },
            # },
            "word": {
                "embedding_dim": 2,
                "lowercase_tokens": True,
                "trainable": True,
                "embedder": {"padding_index": 0},
            },
            # "token_characters": {
            #     "indexer": {
            #         "type": "characters",
            #         "character_tokenizer": {"lowercase_characters": True},
            #     },
            #     "embedder": {
            #         "type": "character_encoding",
            #         "embedding": {"embedding_dim": 32, "padding_index": 0},
            #         "encoder": {
            #             "type": "gru",
            #             "input_size": 32,
            #             "hidden_size": 112,
            #             "num_layers": 1,
            #             "bidirectional": True,
            #         },
            #         "dropout": 0.2,
            #     },
            # },
            "char": {
                "embedding_dim": 2,
                "dropout": 0.1,
                "encoder": {
                    "type": "gru",
                    "hidden_size": 2,
                    "num_layers": 1,
                    "bidirectional": True,
                },
                "indexer": {"character_tokenizer": {"lowercase_characters": True}},
                "embedder": {"embedding": {"padding_index": 0}},
            },
        },
        "head": {
            "type": "RecordPairClassification",
            "labels": ["duplicate", "not_duplicate"],
            "dropout": 0.1,
            "field_encoder": {
                "type": "gru",
                "bidirectional": False,
                # "input_size": 4,
                "hidden_size": 4,
                "num_layers": 1,
            },
            "record_encoder": {
                "type": "gru",
                "bidirectional": True,
                # "input_size": 4,
                "hidden_size": 2,
                "num_layers": 1,
            },
            "matcher_forward": {
                "is_forward": True,
                # "hidden_dim": 2,
                "num_perspectives": 10,
                "with_full_match": False,
            },
            "matcher_backward": {
                "is_forward": False,
                # "hidden_dim": 2,
                "num_perspectives": 10,
            },
            "aggregator": {
                "type": "gru",
                "bidirectional": True,
                # "input_size": ??,
                "hidden_size": 2,
                "num_layers": 1,
            },
            "classifier_feedforward": {
                # "input_dim": 8,
                "num_layers": 1,
                "hidden_dims": [4],
                "activations": ["relu"],
                "dropout": [0.1],
            },
            "initializer": {
                "regexes": [
                    ["_output_layer.weight", {"type": "xavier_normal"}],
                    ["_output_layer.bias", {"type": "constant", "val": 0}],
                    [".*linear_layers.*weight", {"type": "xavier_normal"}],
                    [".*linear_layers.*bias", {"type": "constant", "val": 0}],
                    [".*weight_ih.*", {"type": "xavier_normal"}],
                    [".*weight_hh.*", {"type": "orthogonal"}],
                    [".*bias.*", {"type": "constant", "val": 0}],
                    [".*matcher.*match_weights.*", {"type": "kaiming_normal"}],
                ]
            },
        },
    }

    return pipeline_dict


@pytest.fixture
def trainer_config() -> TrainerConfiguration:
    return TrainerConfiguration(
        max_epochs=1,
        optimizer={"type": "adam", "amsgrad": True, "lr": 0.002},
        gpus=0,
    )


def test_train(pipeline_dict, training_dataset, trainer_config, tmp_path):
    """Testing the correct working of prediction, vocab creating and training"""
    pipeline = Pipeline.from_config(pipeline_dict)
    pipeline.predict(record1={"first_name": "Hans"}, record2={"first_name": "Hansel"})

    trainer = Trainer(
        pipeline=pipeline,
        train_dataset=training_dataset,
        valid_dataset=training_dataset,
        trainer_config=trainer_config,
    )
    trainer.fit(tmp_path / "record_bimpm_experiment")


def test_attributions(pipeline_dict, training_dataset):
    pipeline = Pipeline.from_config(pipeline_dict)
    instance = pipeline.head.featurize(
        training_dataset["record1"][0], training_dataset["record2"][0]
    )
    pipeline.model.eval()
    forward_output = pipeline.model.forward_on_instances([instance])

    attributions = pipeline.head._compute_attributions(forward_output[0], instance)

    assert all([isinstance(attribution, Attribution) for attribution in attributions])
    assert len(attributions) == 4
    assert all([isinstance(attr.attribution, float) for attr in attributions])
    assert all([attributions[i].field == "record1" for i in [0, 1]])
    assert all([attributions[i].field == "record2" for i in [2, 3]])
    assert attributions[1].start == 0 and attributions[1].end == 16

    assert attributions[0].text == "@first_name Hans"
    assert attributions[3].text == "@last_name Petre"

    # Raise error when records with different number of record fields
    instance = pipeline.head.featurize(
        record1={"first_name": "Hans", "last_name": "Zimmermann"},
        record2={"first_name": "Hansel"},
    )
    forward_output = pipeline._model.forward_on_instances([instance])

    with pytest.raises(RuntimeError):
        pipeline.head._compute_attributions(forward_output[0], instance)



import contextlib
import logging
import pickle
from pathlib import Path
from typing import TYPE_CHECKING
from typing import Any
from typing import Dict
from typing import Iterable
from typing import Iterator
from typing import List
from typing import Optional
from typing import Tuple
from typing import Union
from typing import cast

import datasets
from allennlp import __version__ as allennlp__version__
from allennlp.data import Instance
from allennlp.data import Vocabulary
from datasets.fingerprint import Hasher
from spacy import __version__ as spacy__version__
from torch.utils.data import Dataset as TorchDataset
from torch.utils.data import IterableDataset as TorchIterableDataset
from tqdm.auto import tqdm

from biome.text import __version__ as biome__version__
from biome.text.featurizer import FeaturizeError
from biome.text.helpers import copy_sign_and_docs

if TYPE_CHECKING:
    import pandas

    from biome.text.pipeline import Pipeline


class AllennlpDataset(TorchDataset):
    """A thin wrapper around a list of instances.

    Parameters
    ----------
    instances
        List of instances
    vocab
        An optional vocab. This can also be set later with the `.index_with` method.
    """

    def __init__(self, instances: List[Instance], vocab: Optional[Vocabulary] = None):
        self.instances = instances
        self.vocab = vocab

    def __getitem__(self, idx) -> Instance:
        if self.vocab is not None:
            self.instances[idx].index_fields(self.vocab)
        return self.instances[idx]

    def __len__(self):
        return len(self.instances)

    def __iter__(self) -> Iterator[Instance]:
        """
        Even though it's not necessary to implement this because Python can infer
        this method from `__len__` and `__getitem__`, this helps with type-checking
        since `AllennlpDataset` can be considered an `Iterable[Instance]`.
        """
        yield from self.instances

    def index_with(self, vocab: Vocabulary):
        self.vocab = vocab


class AllennlpLazyDataset(TorchIterableDataset):
    """A thin wrapper around a generator of instances.

    Parameters
    ----------
    instance_generator
        A generator that yields instances.
    vocab
        An optional vocab. This can also be set later with the `.index_with` method.
    """

    def __init__(
        self,
        instance_generator: Iterable[Instance],
        vocab: Vocabulary = None,
    ) -> None:
        super().__init__()
        self._instance_generator = instance_generator
        self.vocab = vocab

    def __iter__(self) -> Iterator[Instance]:
        for instance in self._instance_generator:
            if self.vocab is not None:
                instance.index_fields(self.vocab)
            yield instance

    def index_with(self, vocab: Vocabulary):
        self.vocab = vocab


InstanceDataset = Union[AllennlpDataset, AllennlpLazyDataset]


class Dataset:
    """A dataset to be used with biome.text Pipelines

    Is is a very light wrapper around HuggingFace's awesome `datasets.Dataset`,
    only including a biome.text specific `to_instances` method.

    Most of the `datasets.Dataset` API is exposed and can be looked up in detail here:
    https://huggingface.co/docs/datasets/master/package_reference/main_classes.html#datasets.Dataset

    Parameters
    ----------
    dataset
        A HuggingFace `datasets.Dataset`

    Attributes
    ----------
    dataset
        The underlying HuggingFace `datasets.Dataset`
    """

    _LOGGER = logging.getLogger(__name__)
    _CACHED_INSTANCE_LIST_EXTENSION = "instance_list"

    def __init__(self, dataset: datasets.Dataset):
        self.dataset: datasets.Dataset = dataset

    @classmethod
    def load_dataset(cls, *args, split, **kwargs) -> "Dataset":
        """Load a dataset using Huggingface's `datasets.load_dataset` method.

        See https://huggingface.co/docs/datasets/loading_datasets.html

        Parameters
        ----------
        split
            See https://huggingface.co/docs/datasets/splits.html
        *args/**kwargs
            Passed on to the `datasets.load_dataset` method

        Returns
        -------
        dataset
        """
        return cls(datasets.load_dataset(*args, split=split, **kwargs))

    @classmethod
    def from_json(cls, paths: Union[str, List[str]], **kwargs) -> "Dataset":
        """Create a Dataset from a json file

        Parameters
        ----------
        paths
            One or several paths to json files
        **kwargs
            Passed on to the `load_dataset` method

        Returns
        -------
        dataset
        """
        return cls.load_dataset("json", data_files=paths, split="train", **kwargs)

    @classmethod
    def from_csv(cls, paths: Union[str, List[str]], **kwargs) -> "Dataset":
        """Create a Dataset from a csv file

        Parameters
        ----------
        paths
            One or several paths to csv files
        **kwargs
            Passed on to the `load_dataset` method

        Returns
        -------
        dataset
        """
        return cls.load_dataset("csv", data_files=paths, split="train", **kwargs)

    @classmethod
    @copy_sign_and_docs(datasets.Dataset.from_pandas)
    def from_pandas(cls, *args, **kwargs):
        """
        https://huggingface.co/docs/datasets/master/package_reference/main_classes.html#datasets.Dataset.from_pandas
        """
        return cls(datasets.Dataset.from_pandas(*args, **kwargs))

    @classmethod
    @copy_sign_and_docs(datasets.Dataset.from_dict)
    def from_dict(cls, *args, **kwargs) -> "Dataset":
        """
        https://huggingface.co/docs/datasets/master/package_reference/main_classes.html#datasets.Dataset.from_dict
        """
        return cls(datasets.Dataset.from_dict(*args, **kwargs))

    @classmethod
    def from_datasets(cls, dataset_list: List["Dataset"]) -> "Dataset":
        """Create a single Dataset by concatenating a list of datasets

        Parameters
        ----------
        dataset_list
            Datasets to be concatenated. They must have the same column types.

        Returns
        -------
        dataset
        """
        return cls(datasets.concatenate_datasets([ds.dataset for ds in dataset_list]))

    @classmethod
    @copy_sign_and_docs(datasets.Dataset.load_from_disk)
    def load_from_disk(cls, *args, **kwargs) -> "Dataset":
        """
        https://huggingface.co/docs/datasets/master/package_reference/main_classes.html#datasets.Dataset.load_from_disk
        """
        return cls(datasets.load_from_disk(*args, **kwargs))

    @copy_sign_and_docs(datasets.Dataset.save_to_disk)
    def save_to_disk(self, dataset_path: str):
        """
        https://huggingface.co/docs/datasets/master/package_reference/main_classes.html#datasets.Dataset.save_to_disk
        """
        self.dataset.save_to_disk(dataset_path=dataset_path)

    @copy_sign_and_docs(datasets.Dataset.select)
    def select(self, *args, **kwargs) -> "Dataset":
        """
        https://huggingface.co/docs/datasets/master/package_reference/main_classes.html#datasets.Dataset.select
        """
        return Dataset(self.dataset.select(*args, **kwargs))

    @copy_sign_and_docs(datasets.Dataset.map)
    def map(self, *args, **kwargs) -> "Dataset":
        """
        https://huggingface.co/docs/datasets/master/package_reference/main_classes.html#datasets.Dataset.map
        """
        return Dataset(self.dataset.map(*args, **kwargs))

    @copy_sign_and_docs(datasets.Dataset.filter)
    def filter(self, *args, **kwargs) -> "Dataset":
        """
        https://huggingface.co/docs/datasets/master/package_reference/main_classes.html#datasets.Dataset.filter
        """
        return Dataset(self.dataset.filter(*args, **kwargs))

    @copy_sign_and_docs(datasets.Dataset.flatten_)
    def flatten_(self, *args, **kwargs):
        """
        https://huggingface.co/docs/datasets/master/package_reference/main_classes.html#datasets.Dataset.flatten_
        """
        self.dataset.flatten_(*args, **kwargs)

    @copy_sign_and_docs(datasets.Dataset.rename_column_)
    def rename_column_(self, *args, **kwargs):
        """
        https://huggingface.co/docs/datasets/master/package_reference/main_classes.html#datasets.Dataset.rename_column_
        """
        self.dataset.rename_column_(*args, **kwargs)

    @copy_sign_and_docs(datasets.Dataset.remove_columns_)
    def remove_columns_(self, *args, **kwargs):
        """
        https://huggingface.co/docs/datasets/master/package_reference/main_classes.html#datasets.Dataset.remove_columns_
        """
        self.dataset.remove_columns_(*args, **kwargs)

    @copy_sign_and_docs(datasets.Dataset.shuffle)
    def shuffle(self, *args, **kwargs) -> "Dataset":
        """
        https://huggingface.co/docs/datasets/master/package_reference/main_classes.html#datasets.Dataset.shuffle
        """
        return Dataset(self.dataset.shuffle(*args, **kwargs))

    @copy_sign_and_docs(datasets.Dataset.sort)
    def sort(self, *args, **kwargs) -> "Dataset":
        """
        https://huggingface.co/docs/datasets/master/package_reference/main_classes.html#datasets.Dataset.sort
        """
        return Dataset(self.dataset.sort(*args, **kwargs))

    @copy_sign_and_docs(datasets.Dataset.train_test_split)
    def train_test_split(self, *args, **kwargs) -> Dict[str, "Dataset"]:
        """
        https://huggingface.co/docs/datasets/master/package_reference/main_classes.html#datasets.Dataset.train_test_split
        """
        dataset_dict = self.dataset.train_test_split(*args, **kwargs)
        return {
            "train": Dataset(dataset_dict["train"]),
            "test": Dataset(dataset_dict["test"]),
        }

    @copy_sign_and_docs(datasets.Dataset.unique)
    def unique(self, *args, **kwargs) -> List[Any]:
        """
        https://huggingface.co/docs/datasets/master/package_reference/main_classes.html#datasets.Dataset.unique
        """
        return self.dataset.unique(*args, **kwargs)

    @copy_sign_and_docs(datasets.Dataset.set_format)
    def set_format(self, *args, **kwargs):
        """
        https://huggingface.co/docs/datasets/master/package_reference/main_classes.html#datasets.Dataset.set_format
        """
        self.dataset.set_format(*args, **kwargs)

    @copy_sign_and_docs(datasets.Dataset.reset_format)
    def reset_format(self):
        """
        https://huggingface.co/docs/datasets/master/package_reference/main_classes.html#datasets.Dataset.reset_format
        """
        self.dataset.reset_format()

    @contextlib.contextmanager
    @copy_sign_and_docs(datasets.Dataset.formatted_as)
    def formatted_as(self, *args, **kwargs):
        """
        https://huggingface.co/docs/datasets/master/package_reference/main_classes.html#datasets.DatasetDict.formatted_as
        """
        with self.dataset.formatted_as(*args, **kwargs) as internal_cm:
            try:
                yield internal_cm
            finally:
                pass

    @property
    def column_names(self) -> List[str]:
        """Names of the columns in the dataset"""
        return self.dataset.column_names

    @property
    def shape(self) -> Tuple[int]:
        """Shape of the dataset (number of columns, number of rows)"""
        return self.dataset.shape

    @property
    def num_columns(self) -> int:
        """Number of columns in the dataset"""
        return self.dataset.num_columns

    @property
    def num_rows(self) -> int:
        """Number of rows in the dataset (same as `len(dataset)`)"""
        return self.dataset.num_rows

    @property
    def format(self) -> dict:
        return self.dataset.format

    def to_instances(
        self,
        pipeline: "Pipeline",
        lazy: bool = False,
        use_cache: bool = True,
        tqdm_desc: Optional[str] = "Loading instances",
        disable_tqdm: bool = False,
    ) -> InstanceDataset:
        """Convert input to instances for the pipeline

        Parameters
        ----------
        pipeline
            The pipeline for which to create the instances.
        lazy
            If True, instances are lazily loaded from disk, otherwise they are loaded into memory.
        use_cache
            If True, we will try to reuse cached instances. Ignored when `lazy=True`.
        tqdm_desc
            Description for the tqdm progress bar. Default: "Loading instances".
        disable_tqdm
            If True, disable the tqdm progress bar. Default: False

        Returns
        -------
        instance_dataset
        """
        input_columns = [(col, False) for col in pipeline.inputs if col]
        input_columns.extend([(col, True) for col in pipeline.output if col])

        if lazy:
            return AllennlpLazyDataset(
                instance_generator=self._build_instance_generator(
                    pipeline, self.dataset, input_columns
                ),
                vocab=pipeline.vocab,
            )

        fingerprint = self._create_fingerprint_for_instance_list(pipeline)
        instance_list = self._load_instance_list(fingerprint) if use_cache else None
        if instance_list is None:
            instance_generator = self._build_instance_generator(
                pipeline, self.dataset, input_columns
            )
            tqdm_prog = tqdm(
                instance_generator,
                desc=tqdm_desc,
                total=len(self.dataset),
                disable=disable_tqdm,
            )
            instance_list = [instance for instance in tqdm_prog]
            self._cache_instance_list(instance_list, fingerprint)

        return AllennlpDataset(instance_list, vocab=pipeline.vocab)

    def _build_instance_generator(
        self,
        pipeline: "Pipeline",
        dataset: datasets.Dataset,
        input_columns: List[Tuple[str, bool]],
    ) -> Iterable[Instance]:
        """Build the instance generator

        Parameters
        ----------
        pipeline
            The pipeline for which the instances are created for
        dataset
            The dataset underlying the instances
        input_columns
            The columns in the dataset used for the instances

        Returns
        -------
        Iterable[Instance]
        """

        def instance_generator() -> Iterable[Instance]:
            for row in dataset:
                try:
                    instance = pipeline.head.featurize(
                        **{
                            key: row.get(key) if optional else row[key]
                            for key, optional in input_columns
                        }
                    )
                except FeaturizeError as error:
                    self._LOGGER.warning(error)
                else:
                    yield instance

        return instance_generator()

    def _create_fingerprint_for_instance_list(self, pipeline: "Pipeline") -> str:
        """Create a fingerprint for the instance list

        The fingerprint is based on:
        - the fingerprint of the previous dataset
        - the tokenizer config
        - the indexer config of the features
        - the biome__version__, allennlp__version__ and spaCy__version__ just to be completely sure!

        Parameters
        ----------
        pipeline
            Pipeline with the tokenizer and indexer config of the features

        Returns
        -------
        fingerprint
            String of hexadecimal digits
        """
        hasher = Hasher()
        hasher.update(self.dataset._fingerprint)  # necessary evil ...
        hasher.update(vars(pipeline.backbone.tokenizer.config))
        for feature in pipeline.config.features:
            hasher.update(feature.config["indexer"])
        hasher.update(biome__version__)
        hasher.update(allennlp__version__)
        hasher.update(spacy__version__)

        return hasher.hexdigest()

    def _load_instance_list(self, fingerprint: str) -> Optional[List[Instance]]:
        """Load the cached instance list

        Parameters
        ----------
        fingerprint
            Fingerprint of the instance list

        Returns
        -------
        instance_list
            Returns None, if no cached instances are found
        """
        try:
            cache_path = (
                Path(self.dataset.cache_files[0]["filename"]).parent
                / f"{fingerprint}.{self._CACHED_INSTANCE_LIST_EXTENSION}"
            )
            with cache_path.open("rb") as file:
                self._LOGGER.warning(f"Reusing cached instances ({cache_path})")
                instance_list = pickle.load(file)
        except (IndexError, KeyError, FileNotFoundError):
            return None

        return instance_list

    def _cache_instance_list(self, instance_list: List[Instance], fingerprint: str):
        """Cache an instance list

        Parameters
        ----------
        instance_list
            List of instances to be cached
        fingerprint
            Fingerprint of the instance list
        """
        try:
            cache_path = (
                Path(self.dataset.cache_files[0]["filename"]).parent
                / f"{fingerprint}.{self._CACHED_INSTANCE_LIST_EXTENSION}"
            )
            with cache_path.open("xb") as file:
                self._LOGGER.info(f"Caching instances to {cache_path})")
                pickle.dump(instance_list, file)
        except (IndexError, KeyError, FileNotFoundError, FileExistsError):
            pass

    @copy_sign_and_docs(datasets.Dataset.cleanup_cache_files)
    def cleanup_cache_files(self):
        # Apart from cleaning up datasets cache files, this also checks if there is actually a cache dir
        nr_of_removed_files = self.dataset.cleanup_cache_files()
        if nr_of_removed_files is not None:
            cache_dir_path = Path(self.dataset.cache_files[0]["filename"]).parent
            # cleaning up the cached instance lists
            for file_path in cache_dir_path.glob(
                f"*.{self._CACHED_INSTANCE_LIST_EXTENSION}"
            ):
                file_path.unlink()
                self._LOGGER.info(f"Removed {file_path}")
                nr_of_removed_files += 1

        return nr_of_removed_files

    def head(self, n: Optional[int] = 10) -> "pandas.DataFrame":
        """Return the first n rows of the dataset as a pandas.DataFrame

        Parameters
        ----------
        n
            Number of rows. If None, return the whole dataset as a pandas DataFrame

        Returns
        -------
        dataframe
        """
        with self.dataset.formatted_as("pandas"):
            return cast("pandas.DataFrame", self.dataset[:n])

    def __del__(self):
        self.dataset.__del__()

    def __getitem__(self, key: Union[int, slice, str]) -> Union[Dict, List]:
        return self.dataset.__getitem__(key)

    def __len__(self):
        return self.dataset.__len__()

    def __iter__(self):
        yield from self.dataset.__iter__()

    def __repr__(self):
        return self.dataset.__repr__()



from pathlib import Path
from typing import Any
from typing import Dict

import pytorch_lightning
import torch.nn as nn
from allennlp.common import Params
from allennlp.training.learning_rate_schedulers import LearningRateScheduler
from allennlp.training.optimizers import Optimizer

from biome.text import Dataset
from biome.text import Pipeline
from biome.text import Trainer
from biome.text import TrainerConfiguration
from biome.text import VocabularyConfiguration


def _read_configs(configurations_path: Path, section: str) -> Dict[str, Any]:
    code_blocks = {}
    with configurations_path.open() as file:
        in_section = False
        in_new_config = False

        for line in file.readlines():
            if line.startswith(f"## {section}"):
                in_section = True
            elif line.startswith("### ") and in_section:
                code_blocks[line.split(maxsplit=1)[1]] = ""
            elif line.startswith("```python") and in_section:
                in_new_config = True
            elif line.startswith("```") and in_new_config:
                in_new_config = False
            elif line.startswith("## ") and in_section:
                in_section = False

            elif in_section and in_new_config:
                key = list(code_blocks.keys())[-1]
                code_blocks[key] += line

    configurations = {}
    for name, code in code_blocks.items():
        config = {}
        exec(code, globals(), config)
        configurations[name] = config[list(config.keys())[-1]]

    return configurations


def test_pipeline_configs(configurations_path):
    configs = _read_configs(configurations_path, "Pipeline")
    for config_name, config in configs.items():
        Pipeline.from_config(config)


def test_trainer_configs(configurations_path):
    configs = _read_configs(configurations_path, "Trainer")
    pipeline = Pipeline.from_config(
        {
            "name": "test",
            "head": {"type": "TextClassification", "labels": ["pos", "neg"]},
        }
    )
    dataset = Dataset.from_dict({"text": ["test"], "label": ["pos"]})
    linear = nn.Linear(2, 2)
    for config_name, config in configs.items():
        assert isinstance(config, TrainerConfiguration)

        trainer = Trainer(
            pipeline=pipeline, train_dataset=dataset, trainer_config=config
        )
        assert isinstance(trainer.trainer, pytorch_lightning.Trainer)


def test_vocab_configs(configurations_path):
    configs = _read_configs(configurations_path, "Vocabulary")
    for config_name, config in configs.items():
        assert isinstance(config, VocabularyConfiguration)



import pytest
from allennlp.data import Instance
from allennlp.data.fields import ListField
from allennlp.data.fields import TextField
from spacy.tokenizer import Tokenizer
from spacy.vocab import Vocab

from biome.text.configuration import PredictionConfiguration
from biome.text.helpers import spacy_to_allennlp_token
from biome.text.modules.heads import TaskHead
from biome.text.modules.heads import TaskPrediction
from biome.text.modules.heads.task_prediction import Token


@pytest.fixture
def task_head() -> TaskHead:
    return TaskHead(backbone="mock_backbone")


def test_prediction_not_implemented(task_head):
    with pytest.raises(NotImplementedError):
        task_head.make_task_prediction("mock", "mock", "mock")


def test_attributions_not_implemented(task_head, monkeypatch):
    def mock_make_task_prediction(*args, **kwargs):
        return TaskPrediction()

    monkeypatch.setattr(task_head, "_make_task_prediction", mock_make_task_prediction)

    with pytest.raises(NotImplementedError):
        task_head.make_task_prediction(
            "mock", "mock", PredictionConfiguration(add_attributions=True)
        )


def test_make_task_prediction(monkeypatch, task_head):
    def mock_make_task_prediction(*args, **kwargs):
        return TaskPrediction()

    def mock_compute_attributions(*args, **kwargs):
        return kwargs

    def mock_extract_tokens(*args, **kwargs):
        return "tokens"

    monkeypatch.setattr(task_head, "_make_task_prediction", mock_make_task_prediction)
    monkeypatch.setattr(task_head, "_compute_attributions", mock_compute_attributions)
    monkeypatch.setattr(task_head, "_extract_tokens", mock_extract_tokens)

    prediction = task_head.make_task_prediction(
        "mock_forward_output",
        "mock_instance",
        PredictionConfiguration(
            add_tokens=True,
            add_attributions=True,
            attributions_kwargs={"test": "kwarg"},
        ),
    )

    assert isinstance(prediction, TaskPrediction)
    assert hasattr(prediction, "tokens") and hasattr(prediction, "attributions")
    assert prediction.tokens == "tokens"
    assert prediction.attributions == {"test": "kwarg"}


@pytest.mark.parametrize("allennlp_tokens", [False, True])
def test_extract_tokens(task_head, allennlp_tokens):
    tokenizer = Tokenizer(Vocab())
    input_tokens = list(tokenizer("test this sentence."))
    if allennlp_tokens:
        input_tokens = [spacy_to_allennlp_token(tok) for tok in input_tokens]

    tf = TextField(input_tokens, None)
    instance = Instance({"test": tf})

    tokens = task_head._extract_tokens(instance)

    assert all([isinstance(tok, Token) for tok in tokens])
    assert all(itok.text == otok.text for itok, otok in zip(input_tokens, tokens))
    assert all(itok.idx == otok.start for itok, otok in zip(input_tokens, tokens))
    if allennlp_tokens:
        assert all(itok.idx_end == otok.end for itok, otok in zip(input_tokens, tokens))
    else:
        assert all(
            itok.idx + len(itok.text) == otok.end
            for itok, otok in zip(input_tokens, tokens)
        )
    assert all([tok.field == "test" for tok in tokens])


def test_extract_tokens_listfield(task_head):
    tokenizer = Tokenizer(Vocab())
    input_tokens = list(tokenizer("test this sentence."))

    tf = TextField(input_tokens, None)
    instance = Instance({"test": ListField([tf, tf])})

    tokens = task_head._extract_tokens(instance)

    assert len(tokens) == 2 and len(tokens[0]) == 3 and len(tokens[1]) == 3
    assert all(
        [all([isinstance(tok, Token) for tok in tf_tokens] for tf_tokens in tokens)]
    )



from .classification.doc_classification import DocumentClassification
from .classification.doc_classification import DocumentClassificationConfiguration
from .classification.record_classification import RecordClassification
from .classification.record_classification import RecordClassificationConfiguration
from .classification.record_pair_classification import RecordPairClassification
from .classification.record_pair_classification import (
    RecordPairClassificationConfiguration,
)
from .classification.relation_classification import RelationClassification
from .classification.relation_classification import RelationClassificationConfiguration
from .classification.text_classification import TextClassification
from .classification.text_classification import TextClassificationConfiguration
from .language_modelling import LanguageModelling
from .language_modelling import LanguageModellingConfiguration
from .task_head import TaskHead
from .task_head import TaskHeadConfiguration
from .task_head import TaskName
from .task_head import TaskPrediction
from .token_classification import TokenClassification
from .token_classification import TokenClassificationConfiguration

for head in [
    TextClassification,
    TokenClassification,
    DocumentClassification,
    RecordClassification,
    LanguageModelling,
    RecordPairClassification,
    RelationClassification,
]:
    head.register(overrides=True)



import pytest
from allennlp.data.token_indexers import PretrainedTransformerIndexer
from allennlp.data.token_indexers import PretrainedTransformerMismatchedIndexer

from biome.text import Pipeline
from biome.text.configuration import TokenizerConfiguration
from biome.text.tokenizer import Tokenizer
from biome.text.tokenizer import TransformersTokenizer


@pytest.fixture
def pipeline_dict(request) -> dict:
    """Pipeline config dict. You need to update the labels!"""
    pipeline_dict = {
        "name": "transformers_tokenizer_test",
        "features": {
            "transformers": {"model_name": "sshleifer/tiny-distilroberta-base"}
        },
        "head": {
            "type": "TextClassification",
            "labels": ["a", "b"],
        },
    }
    return pipeline_dict


def test_pipeline_transformers_tokenizer(pipeline_dict):
    pipeline_dict["tokenizer"] = {"max_sequence_length": 1}
    pl = Pipeline.from_config(pipeline_dict)

    assert pl.config.tokenizer_config.transformers_kwargs == {
        "model_name": "sshleifer/tiny-distilroberta-base"
    }
    assert pl.config.features.transformers.mismatched is False
    assert (
        type(pl.backbone.featurizer.indexer["transformers"])
        is PretrainedTransformerIndexer
    )
    assert type(pl.backbone.tokenizer) is TransformersTokenizer

    # test max_sequence_length, only <s>, t, </s> should survive
    assert (
        len(pl.backbone.tokenizer.tokenize_text("this is a multi token text")[0]) == 3
    )

    assert pl.predict("Test this!")


def test_pipeline_default_tokenizer(pipeline_dict):
    pipeline_dict["features"].update({"word": {"embedding_dim": 2}})
    pl = Pipeline.from_config(pipeline_dict)

    assert pl.config.tokenizer_config == TokenizerConfiguration()
    assert pl.config.features.transformers.mismatched is True
    assert (
        type(pl.backbone.featurizer.indexer["transformers"])
        is PretrainedTransformerMismatchedIndexer
    )
    assert type(pl.backbone.tokenizer) is Tokenizer

    prediction = pl.predict("Test this!")



import mlflow
import pandas as pd


class BiomeTextModel(mlflow.pyfunc.PythonModel):
    """A custom MLflow model with the 'python_function' flavor for biome.text pipelines.

    This class is used by the `Pipeline.to_mlflow()` method.
    """

    ARTIFACT_CONTEXT = "model"

    def __init__(self):
        self.pipeline = None

    def load_context(self, context):
        from biome.text import Pipeline

        self.pipeline = Pipeline.from_pretrained(
            context.artifacts[self.ARTIFACT_CONTEXT]
        )

    def predict(self, context, dataframe: pd.DataFrame):
        batch = dataframe.to_dict(orient="records")
        predictions = self.pipeline.predict(batch=batch)

        return pd.DataFrame(predictions)
